from Acquisition import aq_base
from collections import OrderedDict
from datetime import date
from datetime import datetime
from euphorie import MessageFactory as _
from euphorie.client import survey
from euphorie.client import utils
from euphorie.client import utils as client_utils
from euphorie.client.model import Risk
from euphorie.client.model import Training
from euphorie.content.behaviors.hide_from_training import IHideFromTraining
from json import dumps
from json import loads
from logging import getLogger
from plone import api
from plone.base.utils import safe_text
from plone.memoize.instance import memoize
from plone.memoize.view import memoize as view_memoize
from Products.Five import BrowserView
from random import sample
from random import shuffle
from sqlalchemy.orm.exc import NoResultFound
from z3c.saconfig import Session
from zExceptions import NotFound
from zExceptions import Unauthorized
from zope.interface import implementer
from zope.publisher.interfaces import IPublishTraverse

import markdown


logger = getLogger(__name__)


class TrainingSlide(BrowserView):
    """Template / macro to hold the training slide markup Currently not active
    in default Euphorie."""

    @property
    @view_memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", self.context, self.request)

    @property
    @memoize
    def is_custom(self):
        return "custom-risks" in self.context.zodb_path

    @property
    @memoize
    def item_type(self):
        return self.context.type

    @property
    @memoize
    def zodb_elem(self):
        if self.is_custom:
            return None
        obj = self.context.aq_parent.restrictedTraverse(
            self.context.zodb_path.split("/"), None
        )
        if obj is None:
            logger.warning(
                "Could not traverse to: %r from %r",
                self.context.zodb_path,
                self.context.aq_parent,
            )
        return obj

    @property
    @memoize
    def risk_type(self):
        elem = self.zodb_elem
        if not elem:
            return ""
        return getattr(elem, "type", "")

    @property
    @memoize
    def for_download(self):
        return "for_download" in self.request and self.request["for_download"]

    @property
    @memoize
    def number(self):
        if self.is_custom:
            num_elems = self.context.number.split(".")
            number = ".".join(["Ω"] + num_elems[1:])
        else:
            number = self.context.number
        if self.item_type == "module":
            number = f"{number}.0"
        return number

    @property
    def number_id(self):
        number = self.number.replace(".", "-")
        if self.is_custom:
            number = number.replace("Ω", "omega")
        return number

    @property
    def slide_title(self):
        if self.is_custom and self.item_type == "module":
            return client_utils.get_translated_custom_risks_title(self.request)
        return self.context.title

    @property
    def slide_template(self):
        if self.item_type == "module":
            return "template-module"
        # there seems to be a profile template, don't when that should be used.
        return "template-default"

    @property
    def slide_type(self):
        # Perhaps this will not be needed any more...
        return self.item_type

    @property
    def slide_date(self):
        return date.today().strftime("%Y-%m-%d")

    @property
    def department(self):
        # XXX tbd - maybe in webhelpers?
        return "-"

    @property
    def description(self):
        if self.is_custom:
            return markdown.markdown(
                getattr(self.context, "custom_description", "") or ""
            )
        return getattr(self.zodb_elem, "description", "") or ""

    @property
    def training_notes(self):
        training_notes = getattr(self.context, "comment", "") or ""
        if self.for_download:
            training_notes = markdown.markdown(training_notes)
        if self.webhelpers.check_markup(training_notes):
            return training_notes

    @property
    def measures_in_place(self):
        if self.item_type != "risk":
            return {}
        measures = OrderedDict()
        for measure in list(self.context.in_place_standard_measures) + list(
            self.context.in_place_custom_measures
        ):
            if not measure.action:
                continue
            measures.update(
                {
                    measure.id: {
                        "action": self.webhelpers.get_safe_html(measure.action),
                        "active": measure.used_in_training,
                    }
                }
            )

        return measures

    @property
    def measures_planned(self):
        if self.item_type != "risk":
            return {}
        measures = OrderedDict()
        for measure in list(self.context.standard_measures) + list(
            self.context.custom_measures
        ):
            if not measure.action:
                continue
            measures.update(
                {
                    measure.id: {
                        "action": self.webhelpers.get_safe_html(measure.action),
                        "active": measure.used_in_training,
                    }
                }
            )

        return measures

    @property
    def measures(self):
        measures = self.measures_in_place
        measures.update(self.measures_planned)
        return measures

    @property
    def image(self):
        if self.is_custom:
            if not getattr(self.context, "image_data", None):
                return None
            _view = self.context.__of__(
                self.webhelpers.traversed_session.aq_parent["custom-risks"]
            ).restrictedTraverse("image-display")
            return _view.get_or_create_image_scaled()

        try:
            image = self.zodb_elem.image.data
        except AttributeError:
            image = None

        if image and self.for_download:
            try:
                scales = self.zodb_elem.restrictedTraverse("images", None)
                if scales:
                    if self.item_type == "module":
                        scale_name = "training_title"
                    else:
                        scale_name = "training"
                    scale = scales.scale(fieldname="image", scale=scale_name)
                    if scale and scale.data:
                        image = scale.data.data
            except Exception:
                image = None
                logger.warning(
                    "Image data could not be fetched on %s", self.context.absolute_url()
                )
        return image

    @property
    def image_urls(self):
        urls = []
        if self.is_custom:
            if not getattr(self.context, "image_data", None):
                return None
            urls.append(
                f"{self.webhelpers.traversed_session.absolute_url()}/"
                f"{'/'.join(self.context.short_path)}"
                f"/@@image-display/training_export?name=${self.context.image_filename}"
            )
            return urls
        field_names = ["image"]
        scale_name = "training_title"
        if self.item_type == "risk":
            field_names.extend(["image2", "image3", "image4"])
            scale_name = "training"
        for fname in field_names:
            image = getattr(self.zodb_elem, fname, None)
            if image and image.data:
                urls.append(
                    f"{self.zodb_elem.absolute_url()}/@@images/{fname}/{scale_name}"
                )
        return urls

    def slide_contents(self):
        return {
            "slide_type": self.item_type,
            "slide_template": self.slide_template,
            "measures_in_place": self.measures_in_place,
            "measures_planned": self.measures_planned,
            "measures": self.measures,  # BBB
            "training_notes": self.training_notes,
        }

    @property
    def existing_measures_training(self):
        measures = dict()
        for measure in list(self.context.in_place_standard_measures) + list(
            self.context.in_place_custom_measures
        ):
            measures.update(
                {
                    measure.id: {
                        "action": measure.action,
                        "active": measure.used_in_training,
                    }
                }
            )
        return measures

    @property
    def planned_measures_training(self):
        measures = dict()
        for measure in list(self.context.standard_measures) + list(
            self.context.custom_measures
        ):
            measures.update(
                {
                    measure.id: {
                        "action": measure.action,
                        "active": measure.used_in_training,
                    }
                }
            )
        return measures


class TrainingView(BrowserView, survey._StatusHelper):
    """The view that shows the main-menu Training module."""

    variation_class = "variation-risk-assessment"
    skip_unanswered = False
    for_download = False
    more_menu_contents = []
    heading_measures_in_place = _(
        "label_existing_measures", default="Already implemented measures"
    )
    heading_measures_planned = _("label_planned_measures", default="Planned measures")
    show_slide_byline = True

    @property
    @view_memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", self.context, self.request)

    @property
    @view_memoize
    def session(self):
        """Return the session for this context/request."""
        return self.context.session

    @property
    @memoize
    def timestamp(self):
        return self.session.modified.strftime("%s%M%H%d%m")

    def get_initial_answers(self):
        """Pick a subset of questions, shuffle them, and initialize the answers
        with None."""
        survey = self.webhelpers._survey
        all_questions = survey.listFolderContents(
            {"portal_type": "euphorie.training_question"}
        )
        num_training_questions = min(
            getattr(survey, "num_training_questions", None) or len(all_questions),
            len(all_questions),
        )
        questions = sample(all_questions, k=num_training_questions)
        return {q.getId(): None for q in questions}

    def get_training_by_id(self, training_id):
        """Return the training for this session with the given id."""
        # check the session id to make sure we are not displaying a training in the
        # context of a different session
        session_id = self.webhelpers.session_id
        try:
            return (
                Session.query(Training)
                .filter(Training.session_id == session_id, Training.id == training_id)
                .one()
            )
        except NoResultFound as exc:
            raise NotFound from exc

    @memoize
    def get_or_create_training(self):
        """Return the current user's training for this session."""
        account_id = self.webhelpers.get_current_account().id
        session_id = self.webhelpers.session_id
        try:
            return (
                Session.query(Training)
                .filter(
                    Training.session_id == session_id, Training.account_id == account_id
                )
                .one()
            )
        except NoResultFound:
            pass
        answers = self.get_initial_answers()
        status = "in_progress" if answers else "correct"

        training = Training(
            account_id=account_id,
            session_id=session_id,
            status=status,
            time=datetime.now(),
            answers=dumps(answers),
        )
        Session.add(training)
        return training

    @property
    @view_memoize
    def training_status(self):
        return self.get_or_create_training().status

    @property
    @view_memoize
    def question_intro_url(self):
        survey = self.webhelpers._survey
        if not getattr(aq_base(survey), "enable_test_questions", False):
            return ""
        view_name = "slide_question_success"
        if survey.listFolderContents(
            {"portal_type": "euphorie.training_question"}
        ) and self.training_status not in ("correct", "success"):
            view_name = "slide_question_intro"
        return f"{self.context.absolute_url()}/@@{view_name}"

    @property
    @view_memoize
    def question_ids(self):
        training = self.get_or_create_training()
        answer_history = loads(training.answers)
        return list(answer_history)

    @property
    def enable_training_questions(self):
        """Explicit property that can be overwritten in subpackages."""
        return bool(self.question_intro_url)

    def slicePath(self, path):
        while path:
            yield path[:3].lstrip("0")
            path = path[3:]

    @property
    @view_memoize
    def title_image(self):
        try:
            return self.context.aq_parent.external_site_logo.data
        except AttributeError:
            logger.warning(
                "Image data (logo) could not be fetched on survey  %s",
                self.context.absolute_url(),
            )
            return

    @property
    @view_memoize
    def tool_image_url(self):
        survey = self.context.aq_parent
        if getattr(survey, "image", None):
            return f"{survey.absolute_url()}/@@images/image/large"

    @property
    def logo_url(self):
        logo = self.webhelpers.get_sector_logo
        if logo:
            return f"{self.webhelpers.portal_url}/{logo.url}"
        return f"{self.webhelpers.portal_url}/++resource++euphorie.resources/assets/oira/style/oira-logo-colour.svg"  # noqa: E501

    @property
    @view_memoize
    def slide_data(self):
        modules = self.getModulePaths()
        risks = self.getRisks(modules, skip_unanswered=self.skip_unanswered)
        seen_modules = []
        hidden_modules = []
        data = OrderedDict()
        for module, risk in risks:
            module_path = module.path
            if module_path in hidden_modules:
                continue
            if module_path not in seen_modules:
                # Get the ZODB representation of the module.
                zodb_module = self.webhelpers.traversed_session.restrictedTraverse(
                    module.zodb_path.split("/")
                )
                # Check if the module is allowed in trainings.
                if (
                    getattr(
                        IHideFromTraining(zodb_module, None),
                        "hide_from_training",
                        False,
                    )
                    is True
                ):
                    hidden_modules.append(module_path)
                    continue

                module_in_context = module.__of__(self.webhelpers.traversed_session)
                module_in_context.REQUEST["for_download"] = self.for_download
                _view = module_in_context.restrictedTraverse("training_slide")
                slide_contents = _view.slide_contents()
                data.update(
                    {
                        module_path: {
                            "item": module_in_context,
                            "training_view": _view,
                            "slide_contents": slide_contents,
                        }
                    }
                )
                seen_modules.append(module_path)
            risk_in_context = risk.__of__(self.webhelpers.traversed_session)
            risk_in_context.REQUEST["for_download"] = self.for_download
            _view = risk_in_context.restrictedTraverse("training_slide")
            slide_contents = _view.slide_contents()
            data.update(
                {
                    risk.path: {
                        "item": risk_in_context,
                        "training_view": _view,
                        "slide_contents": slide_contents,
                    }
                }
            )
        return data

    @property
    def slide_total_count(self):
        count = 0
        for data in self.slide_data.values():
            count += 1
            for measure_id in data["slide_contents"]["measures_in_place"]:
                if data["slide_contents"]["measures_in_place"][measure_id]["active"]:
                    count += 1
                    break
            for measure_id in data["slide_contents"]["measures_planned"]:
                if data["slide_contents"]["measures_planned"][measure_id]["active"]:
                    count += 1
                    break
            if (
                data["item"].type != "module"
                and data["slide_contents"]["training_notes"]
            ):
                count += 1
        return count

    def handle_measure_configuration(self, reply):
        session = Session()
        risk = session.query(Risk).filter(Risk.id == reply["risk_id"]).first()
        if not risk:
            return
        # Gather all (database-) ids of the active measures. That means, those
        # measures where the checkboxes are ticked in the training configuration.
        # Remember: a measure that has been deselected (checkbox unticked)
        # does not appear in the REQUEST
        active_measures_in_place = []
        active_measures_planned = []
        for entry in reply:
            if entry.startswith("training-measure-in-place") and entry.find("-") >= 0:
                measure_id = entry.split("-")[-1]
                active_measures_in_place.append(measure_id)
            elif entry.startswith("training-measure-planned") and entry.find("-") >= 0:
                measure_id = entry.split("-")[-1]
                active_measures_planned.append(measure_id)
        # Get the (database-) ids of all measures-in-place / planned measures
        all_in_place_measures = {
            str(measure.id)
            for measure in list(risk.in_place_standard_measures)
            + list(risk.in_place_custom_measures)
        }
        all_planned_measures = {
            str(measure.id)
            for measure in list(risk.standard_measures) + list(risk.custom_measures)
        }
        # Additionally store the (database-) ids of all measures that have been
        # deactivated.
        deselected_in_place_measures = [
            k for k in all_in_place_measures if k not in active_measures_in_place
        ]
        deselected_planned_measures = [
            k for k in all_planned_measures if k not in active_measures_planned
        ]

        changed = False
        if active_measures_in_place:
            session.execute(
                "UPDATE action_plan set used_in_training=true where id in ({ids})".format(  # noqa: E501
                    ids=",".join(active_measures_in_place)
                )
            )
            changed = True
        if deselected_in_place_measures:
            session.execute(
                "UPDATE action_plan set used_in_training=false where id in ({ids})".format(  # noqa: E501
                    ids=",".join(deselected_in_place_measures)
                )
            )
            changed = True
        if active_measures_planned:
            session.execute(
                "UPDATE action_plan set used_in_training=true where id in ({ids})".format(  # noqa: E501
                    ids=",".join(active_measures_planned)
                )
            )
            changed = True
        if deselected_planned_measures:
            session.execute(
                "UPDATE action_plan set used_in_training=false where id in ({ids})".format(  # noqa: E501
                    ids=",".join(deselected_planned_measures)
                )
            )
            changed = True
        if changed:
            self.webhelpers.traversed_session.session.touch()

    def __call__(self):
        if self.webhelpers.redirectOnSurveyUpdate():
            return

        survey = self.webhelpers._survey
        utils.setLanguage(self.request, survey, survey.language)

        if (
            self.request.environ["REQUEST_METHOD"] == "POST"
            and self.webhelpers.can_edit_session
        ):
            reply = self.request.form
            if "risk_id" in reply:
                self.handle_measure_configuration(reply)

        # XXX This is commented because it causes problems on logout, see:
        # - https://github.com/euphorie/Euphorie/issues/475
        #
        # We should come out with a better solution that does not require setting
        # the cache headers here.
        # self.request.RESPONSE.addHeader("Cache-Control", "public,max-age=60")
        return self.index()


class SlideQuestionIntro(TrainingView):
    """The slide that introduces the questions."""

    @property
    @view_memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", self.context, self.request)

    @property
    def survey_title(self):
        return self.webhelpers._survey.title

    def first_question_url(self):
        """Check the questions in the survey and take the first one."""
        if not self.question_ids:
            return ""
        return "{base_url}/@@slide_question/{slide_id}".format(
            base_url=self.context.absolute_url(), slide_id=self.question_ids[0]
        )


@implementer(IPublishTraverse)
class SlideQuestion(SlideQuestionIntro):
    """The view for a question slide, the question id has to be passed as a
    traversal parameter."""

    def publishTraverse(self, request, name):
        self.question_id = name
        return self

    @property
    @view_memoize
    def question(self):
        """The question we want to display."""
        return self.webhelpers._survey[self.question_id]

    @property
    @view_memoize
    def answers(self):
        """Return the randomized answers for this question."""
        question = self.question
        answers = [
            question.right_answer,
            question.wrong_answer_1,
            question.wrong_answer_2,
        ]
        shuffle(answers)
        return answers

    @property
    def progress(self):
        """Return a progress indicator, something like 2/3."""
        idx = self.question_ids.index(self.question_id)
        return f"{idx + 1}/{len(self.question_ids)}"

    @property
    @view_memoize
    def previous_question_id(self):
        idx = self.question_ids.index(self.question_id)
        if idx == 0:
            return
        try:
            return self.question_ids[idx - 1]
        except IndexError:
            pass

    @property
    @view_memoize
    def next_question_id(self):
        idx = self.question_ids.index(self.question_id)
        try:
            return self.question_ids[idx + 1]
        except IndexError:
            pass

    @property
    def next_url(self):
        """Go to next question (if we have one) or to the success or try again
        slides."""
        next_question_id = self.next_question_id
        if next_question_id:
            next_slide = f"slide_question/{next_question_id}"
        elif self.get_or_create_training().status == "correct":
            next_slide = "slide_question_success"
        else:
            next_slide = "slide_question_try_again"
        return f"{self.context.absolute_url()}/@@{next_slide}"

    def initialize_training(self):
        """Initialize the training.

        This is particularly important if the user starts again the
        training after a first attempt
        """
        training = self.get_or_create_training()
        answer_history = loads(training.answers)
        if not all([answer is None for answer in answer_history.values()]):
            answer_history = {q_id: None for q_id in answer_history}
            training.answers = dumps(answer_history)
        training.status = "in_progress" if answer_history else "correct"
        training.time = datetime.now()

    def post(self):
        if not self.previous_question_id:
            self.initialize_training()
        training = self.get_or_create_training()
        answer = safe_text(self.request.form["answer"])
        answer_history = loads(training.answers)
        answer_history[self.question_id] = answer == self.question.right_answer
        training.answers = dumps(answer_history)
        training.time = datetime.now()
        if not self.next_question_id:
            training.status = (
                "correct"
                if all([answer is True for answer in answer_history.values()])
                else "failed"
            )

    def posted(self):
        if self.request.method != "POST":
            return False
        self.post()
        return True

    def validate(self):
        previous_question_id = self.previous_question_id
        if not previous_question_id:
            return
        training = self.get_or_create_training()
        try:
            answers = loads(training.answers)
        except ValueError:
            answers = {}
        if all([answer is None for answer in answers]):
            raise Unauthorized(_("You should start the training from the beginning"))
        if answers.get(previous_question_id) is None:
            raise Unauthorized(_("It seems you missed a slide"))

    def __call__(self):
        self.validate()
        if self.posted():
            return self.request.response.redirect(self.next_url)
        return super().__call__()


class SlideQuestionSuccess(SlideQuestionIntro):
    variation_class = ""

    def post(self):
        pass

    @property
    @view_memoize
    def organisation_logo(self):
        organisation = self.context.session.account.organisation
        if not organisation or not organisation.image_filename:
            return None
        country_url = self.webhelpers.country_obj.absolute_url()
        return (
            f"{country_url}/@@organisation-logo/{organisation.organisation_id}"
            f"?q={organisation.image_filename}"
        )

    def __call__(self):
        training = self.get_or_create_training()
        if training.status not in ("correct", "success"):
            raise Unauthorized("You do not own the certificate")
        if self.request.method == "POST":
            return self.post()
        return super().__call__()


class SlideQuestionTryAgain(SlideQuestionIntro):
    @property
    def failed_questions(self):
        training = self.get_or_create_training()
        try:
            answers = loads(training.answers)
        except ValueError:
            answers = {}
        survey = self.webhelpers._survey
        return [
            survey.get(question_id).title
            for question_id in self.question_ids
            if answers.get(question_id) is False
        ]


class MyTrainingsPortlet(BrowserView):
    columns = "1"
    element_id = "portlet-training"

    @property
    @memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", self.context, self.request)

    @property
    def available(self):
        return self.my_unfinished_trainings or self.my_certificates

    @property
    @memoize
    def my_unfinished_trainings(self):
        account_id = self.webhelpers.get_current_account().id
        return [
            session
            for session in (
                Session.query(Training)
                .filter(Training.account_id == account_id, Training.status != "correct")
                .order_by(Training.time.desc())
                .all()
            )
            if session.session.tool
        ]

    @property
    @memoize
    def my_certificates(self):
        account_id = self.webhelpers.get_current_account().id
        return [
            session
            for session in (
                Session.query(Training)
                .filter(Training.account_id == account_id, Training.status == "correct")
                .order_by(Training.time.desc())
                .all()
            )
            if session.session.tool
        ]

    def get_certificate(self, session):
        traversed_session = session.traversed_session
        # we cannot call the certificate view directly because if the tool was updated
        # the view will try to redirect (redirectOnSurveyUpdate)
        certificate_view = api.content.get_view(
            name="training-certificate-inner",
            context=traversed_session,
            request=self.request,
        )
        return certificate_view.index()



from euphorie.deployment.upgrade.utils import alembic_upgrade_to
from ftw.upgrade import UpgradeStep


class AddConsultancyTable(UpgradeStep):
    """Add consultancy table."""

    def __call__(self):
        alembic_upgrade_to(self.target_version)



from euphorie.client.browser.publish import handleSurveyUnpublish
from euphorie.content.tests.utils import BASIC_SURVEY
from euphorie.content.tests.utils import createSector
from euphorie.content.tests.utils import EMPTY_SURVEY
from euphorie.testing import EuphorieIntegrationTestCase
from OFS.SimpleItem import SimpleItem


class handleSurveyUnpublishTests(EuphorieIntegrationTestCase):
    def setUp(self):
        super().setUp()
        self.loginAsPortalOwner()
        self.client = self.portal.client

    def createSurvey(self):
        from euphorie.client.tests.utils import addSurvey

        addSurvey(self.portal, BASIC_SURVEY)
        return self.portal.sectors["nl"]["ict"]["software-development"]["test-import"]

    def testUnpublishedSurvey(self):
        from euphorie.content.tests.utils import addSurvey

        sector = createSector(self.portal)
        survey = addSurvey(sector, EMPTY_SURVEY)
        handleSurveyUnpublish(survey, None)

    def testPublishedSurvey(self):
        survey = self.createSurvey()
        clientsector = self.portal.client["nl"]["ict"]
        clientsector["other"] = SimpleItem("other")
        handleSurveyUnpublish(survey, None)
        self.assertEqual(self.portal.client["nl"]["ict"].keys(), ["other"])

    def testRemoveEmptySector(self):
        survey = self.createSurvey()
        clientcountry = self.portal.client["nl"]
        clientcountry["other"] = SimpleItem("other")
        handleSurveyUnpublish(survey, None)
        self.assertEqual(self.portal.client["nl"].keys(), ["other"])



from euphorie.client.tests.utils import addSurvey
from euphorie.client.tests.utils import registerUserInClient
from euphorie.testing import EuphorieFunctionalTestCase
from transaction import commit


class SurveyTests(EuphorieFunctionalTestCase):
    def test_policy_gets_high_priority(self):
        # Test for http://code.simplon.biz/tracker/tno-euphorie/ticket/93
        survey = """<sector xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
                      <title>Sector title</title>
                      <survey>
                        <title>Survey title</title>
                        <language>nl</language>
                        <evaluation-optional>false</evaluation-optional>
                        <module optional="false">
                          <title>Policy parent</title>
                          <description>&lt;p&gt;Een module met een beleidsrisico.&lt;/p&gt;</description>
                          <risk type="policy">
                            <title>Policy problem!</title>
                            <problem-description>There is a policy problem.</problem-description>
                            <description>&lt;p&gt;Random text.&lt;/p&gt;</description>
                            <show-not-applicable>false</show-not-applicable>
                          </risk>
                        </module>
                      </survey>
                    </sector>"""  # noqa: E501
        self.loginAsPortalOwner()
        addSurvey(self.portal, survey)
        commit()
        self.request.response.setHeader("X-Theme-Disabled", "1")
        browser = self.get_browser()
        url = self.portal.client.nl["sector-title"]["survey-title"].absolute_url()
        browser.open(url)
        registerUserInClient(browser)
        # We need to manually open the portlet view as the test browser
        # does not handle JavaScript.
        browser.open(
            self.portal.client["nl"].absolute_url() + "/@@portlet-available-tools"
        )
        # Create a new survey session
        browser.getControl(name="survey").value = ["sector-title/survey-title"]
        browser.getForm(action="new-session").submit()
        browser.getControl(name="form.widgets.title").value = "Test session"
        # Start the survey
        browser.getControl(name="form.button.submit").click()
        session_url = browser.url.replace("/@@identification", "")
        # Identify the risk
        # For debugging we may need some of these settings:
        # browser.handleErrors = False
        # browser.followRedirects = False
        # browser.raiseHttpErrors = False
        browser.open("%s/1/1/@@identification" % session_url)
        browser.getControl(name="answer").value = ["no"]
        browser.getControl(name="next", index=1).click()
        # Check priority in action plan
        browser.open("%s/1/1/@@actionplan" % session_url)
        self.assertEqual(browser.getControl(name="priority").value, ["high"])

    def test_top5_gets_high_priority(self):
        # Test for http://code.simplon.biz/tracker/tno-euphorie/ticket/93
        survey = """<sector xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
                      <title>Sector title</title>
                      <survey>
                        <title>Survey title</title>
                        <language>nl</language>
                        <evaluation-optional>false</evaluation-optional>
                        <module optional="false">
                          <title>Top5 parent</title>
                          <description>&lt;p&gt;Een module met een top-5 risico.&lt;/p&gt;</description>
                          <risk type="top5">
                            <title>Top-5 probleem!</title>
                            <problem-description>Er is een top-5 probleem.</problem-description>
                            <description>&lt;p&gt;Zomaar wat tekst.&lt;/p&gt;</description>
                            <show-not-applicable>false</show-not-applicable>
                          </risk>
                        </module>
                      </survey>
                    </sector>"""  # noqa: E501
        self.loginAsPortalOwner()
        addSurvey(self.portal, survey)
        browser = self.get_browser()
        browser.open(
            self.portal.client.nl["sector-title"]["survey-title"].absolute_url()
        )
        registerUserInClient(browser)
        # We need to manually open the portlet view as the test browser
        # does not handle JavaScript.
        browser.open(
            self.portal.client["nl"].absolute_url() + "/@@portlet-available-tools"
        )
        # Create a new survey session
        browser.getControl(name="survey").value = ["sector-title/survey-title"]
        browser.getForm(action="new-session").submit()
        browser.getControl(name="form.widgets.title").value = "Test session"
        # Start the survey
        browser.getControl(name="form.button.submit").click()
        session_url = browser.url.replace("/@@identification", "")
        # Identify the top-5 risk
        browser.open("%s/1/1/@@identification" % session_url)
        browser.getControl(name="answer").value = ["no"]
        browser.getControl(name="next", index=1).click()
        # Check priority in action plan
        browser.open("%s/1/1/@@actionplan" % session_url)
        self.assertEqual(browser.getControl(name="priority").value, "high")

    def test_top5_skipped_in_evaluation(self):
        # Test for http://code.simplon.biz/tracker/euphorie/ticket/105
        survey = """<sector xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
                      <title>Sector title</title>
                      <survey>
                        <title>Survey title</title>
                        <language>nl</language>
                        <evaluation-optional>false</evaluation-optional>
                        <module optional="false">
                          <title>Top5 parent</title>
                          <description>&lt;p&gt;Een module met een top-5 risico.&lt;/p&gt;</description>
                          <risk type="top5">
                            <title>Top-5 probleem!</title>
                            <problem-description>Er is een top-5 probleem.</problem-description>
                            <description>&lt;p&gt;Zomaar wat tekst.&lt;/p&gt;</description>
                            <show-not-applicable>false</show-not-applicable>
                          </risk>
                        </module>
                      </survey>
                    </sector>"""  # noqa: E501
        self.loginAsPortalOwner()
        addSurvey(self.portal, survey)
        browser = self.get_browser()
        browser.open(
            self.portal.client.nl["sector-title"]["survey-title"].absolute_url()
        )
        registerUserInClient(browser)
        # We need to manually open the portlet view as the test browser
        # does not handle JavaScript.
        browser.open(
            self.portal.client["nl"].absolute_url() + "/@@portlet-available-tools"
        )
        # Create a new survey session
        browser.getControl(name="survey").value = ["sector-title/survey-title"]
        browser.getForm(action="new-session").submit()
        browser.getControl(name="form.widgets.title").value = "Test session"
        # Start the survey
        browser.getControl(name="form.button.submit").click()
        session_url = browser.url.replace("/@@identification", "")
        # Identify the top-5 risk
        browser.open("%s/1/1/@@identification" % session_url)
        browser.getControl(name="answer").value = ["no"]
        # No evaluation is necessary
        self.assertIn(
            "De tool heeft automatisch een risico-evaluatie uitgevoerd",
            browser.contents,
        )



from euphorie.htmllaundry.utils import sanitize
from z3c.form.converter import FieldDataConverter
from z3c.form.interfaces import IWidget
from zope.component import adapter
from zope.interface import implementer
from zope.schema import Text
from zope.schema.interfaces import IText


class IHtmlText(IText):
    pass


@implementer(IHtmlText)
class HtmlText(Text):
    """A HTML field. This is similar to a standard Text field, but will
    sanitize all markup passed into it.
    """

    pass


@adapter(IHtmlText, IWidget)
class HtmlDataConverter(FieldDataConverter):
    """z3c.form data convertor for HTML forms. This convertor
    sanitizes all input, guaranteeing simple and valid markup
    as a result.
    """

    def toFieldValue(self, value):
        data = super().toFieldValue(value)
        if data:
            data = sanitize(data)
        return data



from Products.CMFCore.utils import getToolByName
from Products.Five.browser import BrowserView

import logging


log = logging.getLogger(__file__)


class CleanupContent(BrowserView):
    def __call__(self):
        ct = getToolByName(self.context, "portal_catalog")
        output = []
        for brain in ct(portal_type="euphorie.risk"):
            obj = brain.getObject()
            pd = getattr(obj, "problem_description", None)
            if pd and "[info]" in pd:
                path = "/".join(obj.getPhysicalPath())
                msg = "Removing bogus problem description for %s" % path
                log.debug(msg)
                output.append(msg)
                obj.problem_description = ""

        if not output:
            return "No problematic problem descriptions found"

        return "\n".join(output)



from datetime import timedelta
from euphorie.client import model
from euphorie.client.tests.utils import addAccount
from euphorie.client.tests.utils import addSurvey
from euphorie.content.tests.utils import BASIC_SURVEY
from euphorie.testing import EuphorieIntegrationTestCase
from plone import api
from plone.app.event.base import localized_now


class TestDashboard(EuphorieIntegrationTestCase):
    def setUp(self):
        super().setUp()
        self.loginAsPortalOwner()
        addSurvey(self.portal, BASIC_SURVEY)
        self.account = addAccount(password="secret")
        # We setup two surveys, an archived one and one which is not archived
        survey_session = model.SurveySession(
            title="Dummy session 1",
            zodb_path="nl/ict/software-development",
            account=self.account,
        )
        model.Session.add(survey_session)
        survey_session = model.SurveySession(
            title="Dummy session 2",
            zodb_path="nl/ict/software-development",
            account=self.account,
        )
        survey_session.archived = localized_now() - timedelta(days=1)
        model.Session.add(survey_session)

    def test_portlets_available_tools(self):
        country = self.portal.client.nl

        with api.env.adopt_user(user=self.account):
            with self._get_view("portlet-available-tools", country) as view:
                self.assertListEqual(
                    [survey.getId() for survey in view.surveys],
                    ["software-development"],
                )

                view.request.__annotations__.clear()
                with api.env.adopt_user("admin"):
                    another_one = api.content.create(
                        container=country.ict,
                        type="euphorie.survey",
                        title="Another one",
                    )

                self.assertListEqual(
                    [survey.getId() for survey in view.surveys],
                    ["another-one", "software-development"],
                )

                view.request.__annotations__.clear()

                # Check we can filter previews
                another_one.preview = True

                self.assertListEqual(
                    [survey.getId() for survey in view.surveys],
                    ["software-development"],
                )
                another_one.preview = False

                # Check we can filter obsolete
                view.request.__annotations__.clear()
                another_one.obsolete = True

                self.assertListEqual(
                    [survey.getId() for survey in view.surveys],
                    ["software-development"],
                )
                another_one.obsolete = False

                # Check we can filter by language
                view.request.__annotations__.clear()
                view.request.locale.id.language = "nl"
                another_one.language = "en"
                self.assertListEqual(
                    [survey.getId() for survey in view.surveys],
                    ["software-development"],
                )

                view.request.__annotations__.clear()
                another_one.language = "nl"
                self.assertListEqual(
                    [survey.getId() for survey in view.surveys],
                    ["another-one", "software-development"],
                )

                view.request.__annotations__.clear()
                another_one.language = "nl_NL"
                self.assertListEqual(
                    [survey.getId() for survey in view.surveys],
                    ["another-one", "software-development"],
                )

    def test_portlet_my_ras(self):
        country = self.portal.client.nl

        with api.env.adopt_user(user=self.account):
            with self._get_view("portlet-my-ras", country) as view:
                # The portlet by default hides the archived sessions
                self.assertTrue(view.hide_archived)

                self.assertEqual(
                    len(view.sessions_by_organisation[self.account.organisation]), 1
                )

                view.request.__annotations__.clear()

                # To show it we have to make sure the user unchecked a checkbox
                # that by default is marked (we need and empty marker to check this)
                view.request.set("hide_archived_marker", "1")
                self.assertFalse(view.hide_archived)
                self.assertEqual(
                    len(view.sessions_by_organisation[self.account.organisation]), 2
                )

                view.request.__annotations__.clear()

                # If the checbox is marked, the sessions are hidden again
                view.request.set("hide_archived", "1")
                self.assertTrue(view.hide_archived)
                self.assertEqual(
                    len(view.sessions_by_organisation[self.account.organisation]), 1
                )

            with self._get_view("session-browser-sidebar", country) as view:
                self.assertEqual(len(view.leaf_sessions()), 1)



from euphorie.client.tests.utils import addSurvey
from euphorie.client.tests.utils import registerUserInClient
from euphorie.content.tests.utils import BASIC_SURVEY
from euphorie.testing import EuphorieFunctionalTestCase


class RiskTests(EuphorieFunctionalTestCase):
    def testShowFrenchEvaluation(self):
        # Test for http://code.simplon.biz/tracker/tno-euphorie/ticket/150
        self.loginAsPortalOwner()
        addSurvey(self.portal, BASIC_SURVEY)
        survey = self.portal.client.nl["ict"]["software-development"]
        survey.evaluation_algorithm = "french"
        survey["1"]["2"].type = "risk"
        browser = self.get_browser()
        browser.open(survey.absolute_url())
        registerUserInClient(browser)
        # We need to manually open the portlet view as the test browser
        # does not handle JavaScript.
        browser.open(
            self.portal.client["nl"].absolute_url() + "/@@portlet-available-tools"
        )
        # Create a new survey session
        browser.getControl(name="survey").value = ["ict/software-development"]
        browser.getForm(action="new-session").submit()
        browser.getControl(name="form.widgets.title").value = "Sessiøn".encode()  # noqa
        # Start the survey
        browser.getControl(name="form.button.submit").click()
        session_url = browser.url.replace("/@@identification", "")
        # Identify the risk
        browser.open("%s/1/1/@@identification" % session_url)
        browser.getControl(name="answer").value = ["no"]
        # Verify number of options
        self.assertEqual(
            len(tuple(browser.getControl(name="frequency:int").controls)), 4
        )
        self.assertEqual(
            len(tuple(browser.getControl(name="severity:int").controls)), 4
        )
        # Enter some digits
        browser.getControl(name="frequency:int").value = ["7"]
        browser.getControl(name="severity:int").value = ["10"]
        browser.getControl("next").click()
        # Verify the result
        browser.open("%s/1/1/@@actionplan" % session_url)
        self.assertEqual(browser.getControl(name="priority").value, ["high"])

    def XtestPreventEarlyDate(self):
        """Deactivated until we decide what to do about this kind of validation
        error check."""
        # Test for http://code.simplon.biz/tracker/tno-euphorie/ticket/150
        self.loginAsPortalOwner()
        addSurvey(self.portal, BASIC_SURVEY)
        browser = self.get_browser()
        survey_url = self.portal.client.nl["ict"][
            "software-development"
        ].absolute_url()  # noqa: E501
        browser.open(survey_url)
        registerUserInClient(browser)
        # We need to manually open the portlet view as the test browser
        # does not handle JavaScript.
        browser.open(
            self.portal.client["nl"].absolute_url() + "/@@portlet-available-tools"
        )
        # Create a new survey session
        browser.getControl(name="survey").value = ["ict/software-development"]
        browser.getForm(action="new-session").submit()
        browser.getControl(name="form.widgets.title").value = "Sessiøn".encode()  # noqa
        # Start the survey
        browser.getControl(name="form.button.submit").click()
        # Identify the risk
        browser.getControl("next").click()
        browser.getControl(name="answer").value = ["no"]
        browser.getControl("next").click()
        # Move on to the risk's action plan form
        browser.getLink("Create action plan").click()
        browser.getLink("Next").click()
        # Try an early year
        browser.getControl(
            name="measure.action_plan:utf8:ustring:records", index=0
        ).value = "Do something awesome"
        browser.getControl(name="measure.planning_start_day:records", index=0).value = (
            "1"
        )
        browser.getControl(
            name="measure.planning_start_month:records", index=0
        ).value = ["2"]
        browser.getControl(
            name="measure.planning_start_year:records", index=0
        ).value = "3"
        browser.getControl("next").click()
        self.assertEqual(
            browser.url,
            "http://nohost/plone/client/nl/ict/" "software-development/actionplan/1/1",
        )
        self.assertTrue("Please enter a year between 2000 and 2100" in browser.contents)

    def Xtest_do_not_abort_on_far_future(self):
        """Deactivated, since such a far-future date can not be entered any
        more in modern browsers."""
        # Test for http://code.simplon.biz/tracker/tno-euphorie/ticket/150
        self.loginAsPortalOwner()
        addSurvey(self.portal, BASIC_SURVEY)
        browser = self.get_browser()
        survey_url = self.portal.client.nl["ict"]["software-development"].absolute_url()
        browser.open(survey_url)
        registerUserInClient(browser)
        # We need to manually open the portlet view as the test browser
        # does not handle JavaScript.
        browser.open(
            self.portal.client["nl"].absolute_url() + "/@@portlet-available-tools"
        )
        # Create a new survey session
        browser.getControl(name="survey").value = ["ict/software-development"]
        browser.getForm(action="new-session").submit()
        browser.getControl(name="form.widgets.title").value = "Sessiøn".encode()  # noqa
        # Start the survey
        browser.getControl(name="form.button.submit").click()
        # Identify the risk
        browser.getControl("next").click()
        browser.getControl(name="answer").value = ["no"]
        browser.getControl("next").click()
        # Move on to the risk's action plan form
        browser.getLink("Create action plan").click()
        browser.getLink("Next").click()
        # Try an early year
        browser.getControl(
            name="measure.action_plan:utf8:ustring:records", index=0
        ).value = "Do something awesome"
        browser.getControl(name="measure.planning_start:records", index=0).value = (
            "12345/02/01"
        )
        browser.handleErrors = False
        browser.getControl("next").click()
        self.assertEqual(
            browser.url,
            "http://nohost/plone/client/nl/ict/" "software-development/actionplan/1/1",
        )
        self.assertTrue("Please enter a year between 2000 and 2100" in browser.contents)

    def test_set_unknown_answer_if_skipped(self):
        self.loginAsPortalOwner()
        addSurvey(self.portal, BASIC_SURVEY)
        # Register in the client
        browser = self.get_browser()
        survey_url = self.portal.client.nl["ict"][
            "software-development"
        ].absolute_url()  # noqa: E501
        browser.open(survey_url)
        registerUserInClient(browser)
        # We need to manually open the portlet view as the test browser
        # does not handle JavaScript.
        browser.open(
            self.portal.client["nl"].absolute_url() + "/@@portlet-available-tools"
        )
        # Create a new survey session
        browser.getControl(name="survey").value = ["ict/software-development"]
        browser.getForm(action="new-session").submit()
        browser.getControl(name="form.widgets.title").value = "Sessiøn".encode()  # noqa
        # Start the survey
        browser.getControl(name="form.button.submit").click()
        session_url = browser.url.replace("/@@identification", "")
        # Identify the risk
        browser.open("%s/1/1/@@identification" % session_url)
        # No answer should be set on initial view
        self.assertEqual(browser.getControl(name="answer").value, [])
        # Do not give an identification answer
        risk_url = browser.url
        browser.getControl("next").click()
        # Go back and check the new answer
        browser.open(risk_url)
        self.assertTrue('class="current postponed' in browser.contents)



from euphorie.content.tests.utils import addSurvey
from euphorie.content.tests.utils import BASIC_SURVEY
from euphorie.content.tests.utils import createSector
from euphorie.testing import EuphorieFunctionalTestCase
from plone import api
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from unittest import mock


class LibraryTests(EuphorieFunctionalTestCase):
    _default_credentials = {
        "username": SITE_OWNER_NAME,
        "password": SITE_OWNER_PASSWORD,
    }
    library = [
        {
            "title": "Library Sector Title",
            "url": "http://localhost/sector",
            "surveys": [
                {
                    "title": "Library Survey Title",
                    "path": "/portal/sector/survey/version",
                    "url": "http://localhost/portal/sector/survey/version",
                    "portal_type": "euphorie.sector",
                    "children": [
                        {
                            "title": "Library Risk",
                            "number": "1",
                            "path": "/portal/sector/surveys/version/1",
                            "url": "http://localhost/portal/sector/surveys/version/1",
                            "disabled": False,
                            "portal_type": "euphorie.risk",
                            "children": [],
                        }
                    ],
                }
            ],
        }
    ]

    def test_render(self):
        self.loginAsPortalOwner()
        sector = createSector(self.portal)
        with api.env.adopt_user("admin"):
            survey = addSurvey(sector, BASIC_SURVEY)
        browser = self.get_browser(logged_in=True)
        browser.handleErrors = False
        with mock.patch(
            "euphorie.content.browser.library.get_library", return_value=self.library
        ):
            browser.open(survey.absolute_url() + "/@@library")
        assert "Library Sector Title" in browser.contents
        assert "Library Survey Title" in browser.contents
        assert "Library Risk" in browser.contents



"""
Risk
----

Views for the identification and action plan phases.
"""

from Acquisition import aq_chain
from Acquisition import aq_inner
from Acquisition import aq_parent
from euphorie import MessageFactory as _
from euphorie.client import model
from euphorie.client import utils
from euphorie.client.interfaces import CustomRisksModifiedEvent
from euphorie.client.navigation import FindNextQuestion
from euphorie.client.navigation import FindPreviousQuestion
from euphorie.client.navigation import getTreeData
from euphorie.client.subscribers.imagecropping import _initial_size
from euphorie.content.survey import get_tool_type
from euphorie.content.survey import ISurvey
from euphorie.content.utils import IToolTypesInfo
from euphorie.content.utils import parse_scaled_answers
from euphorie.content.utils import ToolTypesInfo
from euphorie.htmllaundry.utils import strip_markup
from io import BytesIO
from plone import api
from plone.base.utils import safe_text
from plone.memoize.instance import memoize
from plone.memoize.view import memoize_contextless
from plone.namedfile import NamedBlobImage
from plone.namedfile.browser import DisplayFile
from plone.scale.scale import scaleImage
from Products.CMFPlone.utils import getAllowedSizes
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from sqlalchemy import and_
from z3c.saconfig import Session
from zope.component import getUtility
from zope.deprecation import deprecate
from zope.event import notify
from zope.publisher.interfaces import NotFound

import datetime
import PIL


IMAGE_CLASS = {0: "", 1: "twelve", 2: "six", 3: "four", 4: "three"}


class RiskBase(BrowserView):
    # Which fields should be skipped? Default are none, i.e. show all
    skip_fields = []
    # What extra style to use for buttons like "Add measure". Default is None.
    style_buttons = None

    def __call__(self):
        pass

    @property
    @memoize
    def delete_confirmation(self):
        return api.portal.translate(
            _(
                "Are you sure you want to delete this measure? This action can "
                "not be reverted."
            )
        )

    @property
    @memoize
    def override_confirmation(self):
        return api.portal.translate(
            _(
                "The current text in the fields 'Action plan', 'Prevention plan' and "
                "'Requirements' of this measure will be overwritten. "
                "This action cannot be reverted. Are you sure you want to continue?"
            )
        )

    @property
    @memoize
    def message_date_after(self):
        return api.portal.translate(
            _(
                "error_validation_after_start_date",
                default="This date must be on or after the start date.",
            )
        )

    @property
    @memoize
    def message_positive_number(self):
        return api.portal.translate(
            _(
                "error_validation_positive_whole_number",
                default="This value must be a positive whole number.",
            )
        )

    @property
    @memoize
    def message_date_before(self):
        return api.portal.translate(
            _(
                "error_validation_before_end_date",
                default="This date must be on or before the end date.",
            )
        )

    @property
    @memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", self.context.aq_parent, self.request)

    @property
    @memoize
    def session(self):
        return self.webhelpers.traversed_session.session

    @property
    @memoize
    def survey(self):
        """This is the survey dexterity object."""
        return self.webhelpers._survey

    @property
    @memoize_contextless
    def tti(self) -> ToolTypesInfo:
        return getUtility(IToolTypesInfo)

    @property
    @memoize
    def my_tool_type(self) -> str:
        return get_tool_type(self.survey)

    @property
    @memoize
    def use_existing_measures(self) -> bool:
        return (
            api.portal.get_registry_record(
                "euphorie.use_existing_measures", default=False
            )
            and self.my_tool_type in self.tti.types_existing_measures
        )

    @property
    @memoize
    def italy_special(self):
        return self.webhelpers.country == "it"

    @property
    def is_custom_risk(self):
        return self.context.is_custom_risk

    @property
    def solutions_provided_by_tool(self):
        """Return all the solutions that are defined for this risk in the
        CMS."""
        return getattr(self.risk, "_solutions", [])

    @property
    @memoize
    def solutions_available_for_action_plan(self):
        """Return those pre-defined solutions that are not already marked as
        being in place.

        Those remaining solutions are available as suggestions in the
        Action Plan
        """
        if self.is_custom_risk:
            return []
        existing_measure_ids = [
            measure.solution_id for measure in self.get_existing_measures()
        ]
        solutions = []
        for solution in self.solutions_provided_by_tool:
            solution_id = solution.id
            if solution_id not in existing_measure_ids:
                solutions.append(
                    {
                        "description": strip_markup(solution.description),
                        "action": getattr(solution, "action", "") or "",
                        "requirements": solution.requirements,
                        "id": solution_id,
                    }
                )
        return solutions

    @memoize
    def get_existing_measures(self):
        return list(self.context.in_place_standard_measures) + list(
            self.context.in_place_custom_measures
        )

    @property
    def active_standard_measures(self):
        if self.is_custom_risk:
            return {}
        return {
            getattr(measure, "solution_id", ""): measure
            for measure in self.context.standard_measures
        }

    @property
    def solutions_condition(self):
        return "condition: not ({})".format(
            " and ".join(
                [
                    "sm-%s" % solution["id"]
                    for solution in self.solutions_available_for_action_plan
                ]
            )
        )

    def extract_plans_from_request(self):
        """Create new ActionPlan objects by parsing the Request."""
        context = aq_inner(self.context)
        form = self.request.form
        context_measures = context.standard_measures + context.custom_measures
        if form.get("handle_training_measures"):
            # Gather all (database-) ids of the active measures. That means, those
            # measures where the checkboxes are ticked in the training configuration.
            # Remember: a measure that has been deselected (checkbox unticked)
            # does not appear in the REQUEST
            active_measures = []
            for entry in form:
                if entry.startswith("training-measure") and entry.find("-") >= 0:
                    measure_id = entry.split("-")[-1]
                    active_measures.append(measure_id)
            # Get the ids of all measures-in-place and map them to the solution-ids
            # Remember: the solution-id is the ZODB-id of the respective solution
            # inside the risk in the CMS. We us it to identify the standard measures.
            # The custom measures will not have a solution-id
            all_measures = {
                str(measure.id): measure.solution_id for measure in context_measures
            }
            # The following 2 mappings will be used if the handling of measures in
            # place is also part of this request (further down)
            # Make a mapping of solution-id to used-in-training state for the
            # standard measures
            saved_solutions = {
                v: k in active_measures for (k, v) in all_measures.items() if v
            }
            # Make a mapping of database-id to used-in-training state for the
            # custom measures
            saved_custom_measures = {
                k: k in active_measures for (k, v) in all_measures.items() if not v
            }
        else:
            # The following 2 mappings will be used if the handling of measures in
            # place is also part of this request (further down)
            # We do not have training configuration in the REQUEST. We need
            # to build the mapping of solution-id to used-in-training state
            # via the information stored in the the database...
            saved_solutions = {
                plan.solution_id: plan.used_in_training
                for plan in self.context.standard_measures
            }
            # ... and also for the custom measures
            saved_custom_measures = {
                str(plan.id): plan.used_in_training
                for plan in self.context.custom_measures
            }
        new_plans = []
        added = 0
        updated = 0
        existing_plans = {}
        form["action_plans"] = []
        for plan in context_measures:
            existing_plans[str(plan.id)] = plan
        for i in range(0, len(form.get("measure", []))):
            measure = dict([p for p in form["measure"][i].items() if p[1].strip()])
            form["action_plans"].append(measure)
            if len(measure):
                plan_type = measure.get("plan_type", "measure_custom")
                if plan_type == "measure_standard":
                    used_in_training = saved_solutions.get(
                        measure.get("solution_id"), True
                    )
                else:
                    used_in_training = saved_custom_measures.get(
                        measure.get("id"), True
                    )
                solution_id = measure.get("solution_id", None)
                if plan_type == "measure_standard" and not form.get(
                    "sm-%s" % solution_id
                ):
                    continue
                action = measure.get("action", "").strip()
                if not self.webhelpers.check_markup(action):
                    continue
                budget = measure.get("budget")
                budget = budget and budget.split(",")[0].split(".")[0]
                p_start = measure.get("planning_start")
                if p_start:
                    try:
                        datetime.datetime.strptime(p_start, "%Y-%m-%d")
                    except ValueError:
                        p_start = None
                p_end = measure.get("planning_end")
                if p_end:
                    try:
                        datetime.datetime.strptime(p_end, "%Y-%m-%d")
                    except ValueError:
                        p_end = None
                if measure.get("id", "-1") in existing_plans:
                    plan = existing_plans[measure.get("id")]
                    if (
                        action != plan.action
                        or measure.get("requirements") != plan.requirements  # noqa
                        or measure.get("responsible") != plan.responsible
                        or (
                            plan.budget
                            and (budget != str(plan.budget))
                            or plan.budget is None
                            and budget
                        )
                        or (
                            (
                                plan.planning_start
                                and p_start != plan.planning_start.strftime("%Y-%m-%d")
                            )
                            or (plan.planning_start is None and p_start)  # noqa
                        )
                        or (
                            (
                                plan.planning_end
                                and p_end != plan.planning_end.strftime("%Y-%m-%d")
                            )
                            or (plan.planning_end is None and p_end)  # noqa
                        )
                    ):
                        updated += 1
                    del existing_plans[measure.get("id")]
                else:
                    added += 1
                new_plans.append(
                    model.ActionPlan(
                        action=self.webhelpers.get_safe_html(measure.get("action")),
                        requirements=measure.get("requirements"),
                        responsible=measure.get("responsible"),
                        budget=budget,
                        planning_start=p_start,
                        planning_end=p_end,
                        plan_type=plan_type,
                        solution_id=solution_id,
                        used_in_training=used_in_training,
                    )
                )
        removed = len(existing_plans)
        changes = True
        if added == 0 and updated == 0 and removed == 0:
            changes = False
        return (new_plans, changes)

    @property
    def action_plan_instruction_text(self):
        if self.get_existing_measures():
            # Case: measures-in-place==true, solutions==true
            if self.solutions_available_for_action_plan:
                return api.portal.translate(
                    _(
                        "action_measures_true_solutions_true",
                        default="Select or describe any further measure to reduce the risk.",  # noqa: E501  # noqa: E501
                    )
                )
            # Case: measures-in-place==true, solutions==false
            else:
                return api.portal.translate(
                    _(
                        "action_measures_true_solutions_false",
                        default="Describe any further measure to reduce the risk.",
                    )
                )
        else:
            # Case: measures-in-place==false, solutions==true
            if self.solutions_available_for_action_plan:
                return api.portal.translate(
                    _(
                        "action_measures_false_solutions_true",
                        default="Select or describe the specific measures required to reduce the risk.",  # noqa: E501
                    )
                )
            # Case: measures-in-place==false, solutions==false
            else:
                return api.portal.translate(
                    _(
                        "action_measures_false_solutions_false",
                        default="Describe the specific measures required to reduce the risk.",  # noqa: E501
                    )
                )

    def _get_next(self, reply):
        _next = reply.get("next", None)
        # In Safari browser we get a list
        if isinstance(_next, list):
            _next = _next.pop()
        return _next

    @property
    def notes_placeholder(self):
        if self.webhelpers.use_training_module:
            return _(
                "placeholder_comment_field_training",
                default="These notes will be visible in the report and the training. "
                "Use it for anything else you might want to write about this risk.",
            )
        else:
            return _(
                "placeholder_comment_field",
                default="These notes will be visible in the report. Use it for "
                "anything else you might want to write about this risk.",
            )

    @property
    @memoize
    def scaled_answers(self):
        """Get values and answers if the scaled_answers field is used.

        This returns a list of dictionaries.
        """
        if not getattr(self.risk, "use_scaled_answer", False):
            return []
        return parse_scaled_answers(self.risk.scaled_answers)


class IdentificationView(RiskBase):
    """A view for displaying a question in the identification phase."""

    default_template = ViewPageTemplateFile("templates/risk_identification.pt")
    custom_risk_template = ViewPageTemplateFile(
        "templates/risk_identification_custom.pt"
    )
    variation_class = "variation-risk-assessment"

    question_filter = None

    # If False, an always_present risk will not have an explanation that no
    # identification is necessary
    # Default value is True, can be overwritten by certain conditions
    show_explanation_on_always_present_risks = True

    # default value for "always present" risks is "no",
    # can be overwritten by certain conditions
    always_present_answer = "no"

    monitored_properties = {
        "scaled_answer": None,
        "identification": None,
        "postponed": None,
        "frequency": None,
        "severity": None,
        "priority": None,
        "comment": "",
        "training_notes": None,
        "custom_description": None,
    }

    @property
    @memoize
    def risk(self):
        if self.is_custom_risk:
            return
        return self.context.aq_parent.aq_parent.restrictedTraverse(
            self.context.zodb_path.split("/")
        )

    @property
    @memoize
    def default_collapsible_sections(self):
        settings = self.webhelpers.content_country_obj
        default_collapsible_sections = getattr(
            settings,
            "risk_default_collapsible_sections",
            ["collapsible_section_information"],
        )
        return default_collapsible_sections

    @memoize
    def get_collapsible_section_state(self, collapsible_section_name):
        return (
            ""
            if f"collapsible_section_{collapsible_section_name}"
            in self.default_collapsible_sections
            else "closed"
        )

    @property
    @memoize
    def next_question(self):
        return FindNextQuestion(
            self.context, dbsession=self.session, filter=self.question_filter
        )

    @property
    @memoize
    def skip_evaluation(self):
        """Default value is False, but it can be tweaked in certain
        conditions."""
        if self.italy_special and (
            (
                self.risk
                and (self.risk.type == "top5" or self.risk.evaluation_method == "fixed")
            )
            or self.is_custom_risk
        ):
            return True
        return False

    @property
    @memoize
    def evaluation_condition(self):
        """In what circumstances will the Evaluation panel be shown, provided
        that evaluation is not skipped in general?

        If you are using scaled answers instead of yes/no, you likely want to
        override this.
        """
        condition = "condition: answer=no"
        if self.italy_special and not self.skip_evaluation:
            condition = "condition: answer=no or answer=yes"
        return condition

    @property
    def action_plan_condition(self):
        """In what circumstances will the integrated Action Plan be shown."""
        condition = "condition: answer=no"
        if not self.is_custom_risk and (
            self.risk.type == "top5" or self.risk.risk_always_present
        ):
            # No condition, that means, it will always be shown
            return None
        return condition

    def __call__(self):
        # Render the page only if the user has inspection rights,
        # otherwise redirect to the start page of the session.
        if not self.webhelpers.can_inspect_session:
            return self.request.response.redirect(
                self.context.aq_parent.absolute_url() + "/@@start"
            )
        super().__call__()
        self.check_render_condition()

        utils.setLanguage(self.request, self.survey, self.survey.language)

        if self.request.method == "POST":
            reply = self.request.form
            if not self.webhelpers.can_edit_session:
                return self.proceed_to_next(reply)
            _next = self._get_next(reply)
            # Don't persist anything if the user skipped the question
            if _next == "skip":
                return self.proceed_to_next(reply)
            old_values = {}
            for prop, default in self.monitored_properties.items():
                val = getattr(self.context, prop, default)
                old_values[prop] = val

            self.set_answer_data(reply)

            session = Session()
            changed = self.set_measure_data(reply, session)

            if reply.get("answer", None):
                # If answer is not present in the request, do not attempt to set
                # any action-related data, since the request might have come
                # from a sub-form.
                if self.webhelpers.integrated_action_plan:
                    new_plans, changes = self.extract_plans_from_request()
                    for plan in (
                        self.context.standard_measures + self.context.custom_measures
                    ):
                        session.delete(plan)
                    self.context.action_plans.extend(new_plans)
                    changed = changes or changed

            # This only happens on custom risks
            if reply.get("handle_custom_description"):
                self.context.custom_description = self.webhelpers.check_markup(
                    reply.get("custom_description")
                )

            if reply.get("title"):
                self.context.title = reply.get("title")

            if not changed:
                for prop, default in self.monitored_properties.items():
                    val = getattr(self.context, prop, None)
                    if val and val != old_values[prop]:
                        changed = True
                        break
            if changed:
                self.session.touch()

            return self.proceed_to_next(reply)

        else:
            self._prepare_risk()
            if self.is_custom_risk:
                next = FindNextQuestion(
                    self.context, self.context.session, filter=self.question_filter
                )
                self.has_next_risk = next or False
            return self.template()

    @property
    def template(self):
        if self.is_custom_risk:
            return self.custom_risk_template
        return self.default_template

    @property
    def tree(self):
        return getTreeData(self.request, self.context)

    @property
    def title(self):
        return self.session.title

    def check_render_condition(self):
        # Render the page only if the user can inspection rights,
        # otherwise redirect to the start page of the session.
        if not self.webhelpers.can_inspect_session:
            return self.request.response.redirect(
                "{session_url}/@@start".format(
                    session_url=self.webhelpers.traversed_session.absolute_url()
                )
            )
        if self.webhelpers.redirectOnSurveyUpdate():
            return

    @deprecate(
        "This method does not do anything anymore. The set attributes have been moved to properties."  # noqa: E501
    )
    def set_parameter_values(self):
        """This method will be removed. It was used to set the values of:

        - tti
        - my_tool_type
        - use_existing_measures

        These values are now properties of the RiskBase class.
        """
        pass

    def get_identification_from_scaled_answer(self, scaled_answer):
        """Determine the yes/no identification based on the scaled answer.

        A simplistic implementation could be:

          return "no" if scaled_answer in ("1", "2") else "yes"

        You likely want to override this if you actually use scaled answers,
        so by default we return nothing, making the identification empty.
        """
        pass

    def set_answer_data(self, reply):
        """Set answer data from the reply.

        For years the only answer possibilities were yes, no, n/a or postponed.
        Now we may have an extra field scaled_answer, for answers in the
        range of (usually) 1-5.
        We might want to merge these two possibilities, but for now they are
        separate.
        Currently, when scaled_answer is in the reply, we also get
        'postponed' as answer.  We can ignore this: scaled_answer is filled
        in, so the answer is not postponed.
        We make sure that either 'identification' is set (yes/no) or
        'scaled_answer' is set (1-5), and the other None.
        Or both None in the case the answer is really postponed.
        """
        answer = reply.get("answer", None)
        # If answer is not present in the request, do not attempt to set
        # any answer-related data, since the request might have come
        # from a sub-form.
        if not answer:
            return
        self.context.comment = self.webhelpers.get_safe_html(reply.get("comment"))
        scaled_answer = reply.get("scaled_answer", None)
        if scaled_answer:
            # We have an answer on the scale of 1-5 (or similar).
            self.context.scaled_answer = scaled_answer
            self.context.postponed = False
            self.context.identification = self.get_identification_from_scaled_answer(
                scaled_answer
            )
        else:
            self.context.scaled_answer = None
            self.context.postponed = answer == "postponed"
            if self.context.postponed:
                self.context.identification = None
                return
            self.context.identification = answer
        if getattr(self.risk, "type", "") in ("top5", "policy"):
            self.context.priority = "high"
        elif getattr(self.risk, "evaluation_method", "") == "calculated":
            self.calculatePriority(self.risk, reply)
        elif self.risk is None or self.risk.evaluation_method == "direct":
            self.context.priority = reply.get("priority")

    def set_measure_data(self, reply, session):
        changed = False
        # Case: the user has selected or de-selected a measure
        # from the training configuration
        if reply.get("handle_training_measures"):
            # Gather all (database-) ids of the active measures. That means, those
            # measures where the checkboxes are ticked in the training configuration.
            # Remember: a measure that has been deselected (checkbox unticked)
            # does not appear in the REQUEST
            active_measures_in_place = []
            active_measures_planned = []
            for entry in reply:
                if (
                    entry.startswith("training-measure-in-place")
                    and entry.find("-") >= 0
                ):
                    measure_id = entry.split("-")[-1]
                    active_measures_in_place.append(measure_id)
                elif (
                    entry.startswith("training-measure-planned")
                    and entry.find("-") >= 0
                ):
                    measure_id = entry.split("-")[-1]
                    active_measures_planned.append(measure_id)
            # Get the ids of all measures-in-place and map them to the solution-ids
            # Remember: the solution-id is the ZODB-id of the respective solution
            # inside the risk in the CMS. We us it to identify the standard measures.
            # The custom measures will not have a solution-id
            all_in_place_measures = {
                str(measure.id): measure.solution_id
                for measure in list(self.context.in_place_standard_measures)
                + list(self.context.in_place_custom_measures)
            }
            all_planned_measures = {
                str(measure.id): measure.solution_id
                for measure in list(self.context.standard_measures)
                + list(self.context.custom_measures)
            }
            # The following 2 mappings will be used if the handling of measures in
            # place is also part of this request (further down)
            # Make a mapping of solution-id to used-in-training state for the
            # standard measures
            saved_in_place_solutions = {
                v: k in active_measures_in_place
                for (k, v) in all_in_place_measures.items()
                if v
            }
            # Make a mapping of database-id to used-in-training state for the
            # custom measures
            saved_in_place_custom_measures = {
                k: k in active_measures_in_place
                for (k, v) in all_in_place_measures.items()
                if not v
            }
            # Additionally store the (database) ids of all measures that have been
            # deactivated. In case we do not handle measures in place in this
            # request further down, we need to make sure that those measues get
            # set to used_in_training=False at the end of this method
            deselected_in_place_measures = [
                k for k in all_in_place_measures if k not in active_measures_in_place
            ]
            deselected_planned_measures = [
                k for k in all_planned_measures if k not in active_measures_planned
            ]
        else:
            # The following 2 mappings will be used if the handling of measures in
            # place is also part of this request (further down)
            # We do not have training configuration in the REQUEST. We need
            # to build the mapping of solution-id to used-in-training state
            # via the information stored in the the database...
            saved_in_place_solutions = {
                plan.solution_id: plan.used_in_training
                for plan in self.context.in_place_standard_measures
            }
            # ... and also for the custom measures
            saved_in_place_custom_measures = {
                str(plan.id): plan.used_in_training
                for plan in self.context.in_place_custom_measures
            }
            deselected_in_place_measures = deselected_planned_measures = []
            active_measures_in_place = active_measures_planned = []

        if reply.get("handle_measures_in_place"):
            new_measures = []
            # First, check which of the standard solutions were selected
            for solution in self.solutions_provided_by_tool:
                # If the solution was already added as a measure, retrieve the info
                # whether it was de-selected for the training, so that we don't
                # forcefully activate all measures for the training again.
                used_in_training = saved_in_place_solutions.get(solution.id, True)
                if reply.get("measure-standard-%s" % solution.id):
                    new_measures.append(
                        model.ActionPlan(
                            action=solution.action,
                            requirements=solution.requirements,
                            plan_type="in_place_standard",
                            solution_id=solution.id,
                            used_in_training=used_in_training,
                        )
                    )
            # Clean up, remove all standard in-place measures. The active
            # ones will be created freshly again below
            for plan in self.context.in_place_standard_measures:
                session.delete(plan)

            # Now loop over the form data to find out which new custom measures
            # to add and which already present custom measures to keep.
            # (Present custom measures are deleted and added again, since their
            # text might have changed. But their original order is preserved.)
            new_custom_measures = {}
            for k, val in reply.items():
                if (
                    k.startswith("new-measure")
                    and isinstance(val, str)
                    and val.strip() != ""
                ):
                    new_custom_measures[k] = model.ActionPlan(
                        action=val, plan_type="in_place_custom"
                    )
                elif k.startswith("measure-custom"):
                    _id = k.rsplit("-", 1)[-1]
                    new_custom_measures[_id] = model.ActionPlan(
                        action=val,
                        plan_type="in_place_custom",
                        used_in_training=saved_in_place_custom_measures.get(_id, True),
                    )
                # This only happens on custom risks
                elif k.startswith("present-measure") and val.strip() != "":
                    _id = k.rsplit("-", 1)[-1]
                    if int(bool(reply.get(f"measure-{_id}"))):
                        new_measures.append(
                            model.ActionPlan(
                                action=val,
                                plan_type="in_place_custom",
                                used_in_training=saved_in_place_custom_measures.get(
                                    _id, True
                                ),
                            )
                        )
            # Now add the custom measures in their correct order
            for k in sorted(new_custom_measures.keys()):
                new_measures.append(new_custom_measures[k])
            # Delete all custom in-place measures
            for plan in self.context.in_place_custom_measures:
                session.delete(plan)
                changed = True

            if new_measures:
                self.context.action_plans.extend(new_measures)
                changed = True
        else:
            # If we did not handle any measures in place, check if we need to take
            # care of training measures that need to be activated or deactivated
            if active_measures_in_place:
                session.execute(
                    "UPDATE action_plan set used_in_training=true where id in ({ids})".format(  # noqa: E501
                        ids=",".join(active_measures_in_place)
                    )
                )
                changed = True
            if deselected_in_place_measures:
                session.execute(
                    "UPDATE action_plan set used_in_training=false where id in ({ids})".format(  # noqa: E501
                        ids=",".join(deselected_in_place_measures)
                    )
                )
                changed = True
        # In case we're not dealing with an integrated action plan, de-/ activate
        # planned measures as necessary
        if not self.webhelpers.integrated_action_plan:
            if active_measures_planned:
                session.execute(
                    "UPDATE action_plan set used_in_training=true where id in ({ids})".format(  # noqa: E501
                        ids=",".join(active_measures_planned)
                    )
                )
                changed = True
            if deselected_planned_measures:
                session.execute(
                    "UPDATE action_plan set used_in_training=false where id in ({ids})".format(  # noqa: E501
                        ids=",".join(deselected_planned_measures)
                    )
                )
                changed = True
        return changed

    @property
    @memoize
    def number_images(self):
        number_images = getattr(self.risk, "image", None) and 1 or 0
        if number_images:
            for i in range(2, 5):
                number_images += getattr(self.risk, f"image{i}", None) and 1 or 0
        return number_images

    def _prepare_risk(self):
        has_risk_description = (
            self.risk and utils.HasText(self.risk.description)
        ) or getattr(self.context, "custom_description", "")
        self.show_info = getattr(self.risk, "image", None) or (
            self.risk is None or utils.HasText(self.risk.description)
        )
        self.image_class = IMAGE_CLASS[self.number_images]
        number_files = 0
        for i in range(1, 5):
            number_files += getattr(self.risk, f"file{i}", None) and 1 or 0
        self.has_files = number_files > 0
        self.has_legal = utils.HasText(getattr(self.risk, "legal_reference", None))
        self.show_resources = self.has_legal or self.has_files

        self.risk_number = self.context.number

        self.description_probability = _(
            "help_default_probability",
            default="Indicate how "
            "likely occurence of this risk is in a normal situation.",
        )
        self.description_frequency = _(
            "help_default_frequency",
            default="Indicate how often this " "risk occurs in a normal situation.",
        )
        self.description_severity = _(
            "help_default_severity",
            default="Indicate the " "severity if this risk occurs.",
        )

        tool_types = self.tti()
        tt_default = self.tti.default_tool_type
        tool_type_data = tool_types.get(self.my_tool_type, tool_types[tt_default])
        default_type_data = tool_types["classic"]
        self.show_existing_measures = False

        # Fill some labels with default texts
        self.answer_yes = tool_type_data.get(
            "answer_yes", default_type_data["answer_yes"]
        )
        self.answer_no = tool_type_data.get("answer_no", default_type_data["answer_no"])
        self.answer_na = tool_type_data.get("answer_na", default_type_data["answer_na"])
        self.intro_extra = ""
        self.intro_questions = tool_type_data.get("intro_questions", "")
        if getattr(self.risk, "risk_always_present", False):
            self.placeholder_add_extra = tool_type_data.get(
                "placeholder_add_extra_always_present", ""
            )
            self.button_add_extra = tool_type_data.get(
                "button_add_extra_always_present", ""
            )
        else:
            self.placeholder_add_extra = tool_type_data.get("placeholder_add_extra", "")
            self.button_add_extra = tool_type_data.get("button_add_extra", "")

        self.button_remove_extra = ""
        if self.use_existing_measures:
            measures = self.get_existing_measures_with_activation()
            # Only show the form to select and add existing measures if
            # at least one pre-existring measure is present
            # In this case, also change some labels
            if len(measures):
                self.show_existing_measures = True
                if getattr(self.risk, "risk_always_present", False):
                    self.intro_extra = tool_type_data.get(
                        "intro_extra_always_present", ""
                    )
                    self.button_remove_extra = tool_type_data.get(
                        "button_remove_extra_always_present", ""
                    )
                else:
                    self.intro_extra = tool_type_data.get("intro_extra", "")
                    self.button_remove_extra = tool_type_data.get(
                        "button_remove_extra", ""
                    )
                if self.webhelpers.integrated_action_plan:
                    self.answer_yes = tool_type_data.get(
                        "answer_yes_integrated_ap", tool_type_data["answer_yes"]
                    )
                    self.answer_no = tool_type_data.get(
                        "answer_no_integrated_ap", tool_type_data["answer_no"]
                    )
                else:
                    self.answer_yes = tool_type_data["answer_yes"]
                    self.answer_no = tool_type_data["answer_no"]
                self.answer_na = tool_type_data["answer_na"]
        if self.is_custom_risk:
            self.intro_extra = tool_type_data.get("custom_intro_extra", "")
            self.intro_questions = tool_type_data.get("custom_intro_questions", "")
            self.button_add_extra = tool_type_data.get("custom_button_add_extra", "")
            self.placeholder_add_extra = tool_type_data.get(
                "custom_placeholder_add_extra", ""
            )

        survey = self.context.aq_parent.aq_parent
        if getattr(survey, "enable_custom_evaluation_descriptions", False):
            if survey.evaluation_algorithm != "french":
                custom_dp = getattr(survey, "description_probability", "") or ""
                self.description_probability = (
                    custom_dp.strip() or self.description_probability
                )
            custom_df = getattr(survey, "description_frequency", "") or ""
            self.description_frequency = custom_df.strip() or self.description_frequency
            custom_ds = getattr(survey, "description_severity", "") or ""
            self.description_severity = custom_ds.strip() or self.description_severity

        # compute training side template
        self.slide_template = (
            (has_risk_description or self.number_images)
            and "template-two-column"
            or "template-default"
        )

    @property
    @memoize
    def previous_question(self):
        return FindPreviousQuestion(
            self.context, dbsession=self.session, filter=self.question_filter
        )

    def proceed_to_next(self, reply):
        _next = reply.get("next", None)
        # In Safari browser we get a list
        if isinstance(_next, list):
            _next = _next.pop()
        if _next == "previous":
            target = self.previous_question
            if target is None:
                # We ran out of questions, step back to intro page
                url = "{session_url}/@@identification".format(
                    session_url=self.webhelpers.traversed_session.absolute_url()
                )
                return self.request.response.redirect(url)
        elif _next in ("next", "skip"):
            target = self.next_question
            if target is None:
                # We ran out of questions, proceed to the action plan
                if self.webhelpers.use_action_plan_phase:
                    next_view_name = "@@actionplan"
                elif self.webhelpers.use_consultancy_phase:
                    next_view_name = "@@consultancy"
                else:
                    next_view_name = "@@report"
                base_url = self.webhelpers.traversed_session.absolute_url()
                url = f"{base_url}/{next_view_name}"
                return self.request.response.redirect(url)

        elif _next == "add_custom_risk" and self.webhelpers.can_edit_session:
            sql_module = (
                Session.query(model.Module)
                .filter(
                    and_(
                        model.SurveyTreeItem.session == self.session,
                        model.Module.zodb_path == "custom-risks",
                    )
                )
                .first()
            )
            if not sql_module:
                url = self.context.absolute_url() + "/@@identification"
                return self.request.response.redirect(url)

            view = api.content.get_view("identification", sql_module, self.request)
            view.add_custom_risk()
            notify(CustomRisksModifiedEvent(self.context.aq_parent))
            risk_id = self.context.aq_parent.children().count()
            # Construct the path to the newly added risk: We know that there
            # is only one custom module, so we can take its id directly. And
            # to that we can append the risk id.
            url = "{session_url}/{module}/{risk}/@@identification".format(
                session_url=self.webhelpers.traversed_session.absolute_url(),
                module=sql_module.getId(),
                risk=risk_id,
            )
            return self.request.response.redirect(url)
        elif _next == "actionplan":
            url = self.webhelpers.traversed_session.absolute_url() + "/@@actionplan"
            return self.request.response.redirect(url)
        # stay on current risk
        else:
            target = self.context
        url = ("{session_url}/{path}/@@identification").format(
            session_url=self.webhelpers.traversed_session.absolute_url(),
            path="/".join(target.short_path),
        )
        return self.request.response.redirect(url)

    @memoize
    def get_existing_measures_with_activation(self):
        saved_standard_measures = {
            getattr(measure, "solution_id", ""): measure
            for measure in self.context.in_place_standard_measures
        }
        existing_measures = []
        for solution in self.solutions_provided_by_tool:
            if getattr(self.survey, "measures_text_handling", "full") == "simple":
                title = solution.action
                text = ""
            else:
                title = solution.description
                text = solution.action
            existing_measures.append(
                {
                    "title": title,
                    "text": text,
                    "active": solution.id in saved_standard_measures,
                    "solution_id": solution.id,
                    "plan_type": "in_place_standard",
                }
            )
        for measure in self.context.in_place_custom_measures:
            existing_measures.append(
                {
                    "text": measure.action,
                    "active": True,
                    "solution_id": measure.id,
                    "plan_type": "in_place_custom",
                }
            )
        return existing_measures

    @property
    def use_problem_description(self):
        text = self.context.problem_description or ""
        return bool(text.strip())

    def evaluation_algorithm(self, risk):
        return evaluation_algorithm(risk)

    def calculatePriority(self, risk, reply):
        self.context.frequency = reply.get("frequency")
        try:
            if evaluation_algorithm(risk) == "french":
                self.context.effect = reply.get("severity")
            else:  # Kinney method
                self.context.effect = reply.get("effect")
                self.context.probability = reply.get("probability")
            calculate_priority(self.context, risk)
        except TypeError:
            pass
        return self.context.priority


class ImageUpload(BrowserView):
    def redirect(self):
        return self.request.response.redirect(
            f"{self.context.absolute_url()}/@@identification"
        )

    @property
    @memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", self.context.aq_parent, self.request)

    def __call__(self):
        if not self.webhelpers.can_view_session:
            # The user cannot call this view to go the sessions overview.
            return self.request.response.redirect(self.webhelpers.client_url)
        if self.request.form.get("image"):
            image = self.request.form["image"]
            new_data = image.read()
            if self.context.image_data != new_data:
                try:
                    pil_img = PIL.Image.open(BytesIO(new_data))
                    pil_img.crop()
                except OSError:
                    api.portal.show_message(
                        _(
                            "Invalid file format for image. Please use PNG, JPEG or GIF."  # noqa: E501
                        ),
                        request=self.request,
                        type="warning",
                    )
                    return self.redirect()
                self.context.image_data = new_data
                self.context.image_data_scaled = None
            new_name = safe_text(image.filename)
            if self.context.image_filename != new_name:
                self.context.image_filename = new_name
        elif self.request.form.get("image-remove"):
            self.context.image_data = None
            self.context.image_filename = ""
        return self.redirect()


class ImageDisplay(DisplayFile):
    """Return the image stored in the risk (if present). Allows also to get the
    image scaled if invoked like:

    ../@@image-display/image_large/${here/image_filename}
    """

    @property
    @memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", self.context.aq_parent, self.request)

    def get_or_create_image_scaled(self):
        """Get the image scaled."""
        if self.context.image_data_scaled:
            return self.context.image_data_scaled
        image = PIL.Image.open(BytesIO(self.context.image_data))
        image_format = image.format or self.DEFAULT_FORMAT
        params = list(image.size)
        scale = getAllowedSizes().get(self.fieldname, (1500, 791))
        params.extend(scale)
        box = _initial_size(*params)

        cropped_image = image.crop(box).resize(scale)
        cropped_image_io = BytesIO()
        cropped_image.save(cropped_image_io, image_format, quality=100)
        scaled_image_io, scaled_image_format, scaled_image_size = scaleImage(
            cropped_image_io, width=scale[0], height=scale[1]
        )
        if isinstance(scaled_image_io, bytes):
            self.context.image_data_scaled = scaled_image_io
        else:
            self.context.image_data_scaled = scaled_image_io.getvalue()
        return self.context.image_data_scaled

    def _getFile(self):
        if self.context.image_data is None:
            raise NotFound(self, self.fieldname, self.request)

        if self.fieldname and "training" in self.fieldname:
            image_data = self.get_or_create_image_scaled()
        else:
            image_data = self.context.image_data

        return NamedBlobImage(image_data, filename=self.context.image_filename)

    def __call__(self):
        if not self.webhelpers.can_view_session:
            # The user cannot call this view to go the sessions overview.
            return self.request.response.redirect(self.webhelpers.client_url)
        return super().__call__()


class ActionPlanView(RiskBase):
    """Logic for creating new action plans."""

    phase = "actionplan"
    variation_class = "variation-risk-assessment"
    # The question filter will find modules AND risks
    question_filter = model.ACTION_PLAN_FILTER
    # The risk filter will only find risks
    risk_filter = model.RISK_PRESENT_OR_TOP5_FILTER

    @property
    @memoize
    def risk(self):
        if self.is_custom_risk:
            return self.context
        return self.context.aq_parent.aq_parent.restrictedTraverse(
            self.context.zodb_path.split("/")
        )

    @property
    @memoize
    def skip_evaluation(self):
        """Default value is False, but it can be tweaked in certain
        conditions."""
        if self.italy_special and (
            (
                self.risk
                and (self.risk.type == "top5" or self.risk.evaluation_method == "fixed")
            )
            or self.is_custom_risk
        ):
            return True
        return False

    @property
    def risk_present(self):
        return self.context.identification == "no"

    @property
    def risk_postponed(self):
        return self.context.identification is None

    @property
    def use_problem_description(self):
        if self.is_custom_risk:
            return False
        if self.italy_special:
            return False
        text = self.risk.problem_description or ""
        return bool(text.strip())

    @property
    def scaled_answer_chosen(self):
        if not self.risk.use_scaled_answer:
            return ""
        if self.context.scaled_answer is None:
            return ""
        answer = self.context.scaled_answer
        # answer is a string like '1'.
        # Use it to find the textual representation of the answer.
        for info in self.scaled_answers:
            # Note: currently both info value and answer are strings ('1', '2', etc).
            if info["value"] == answer:
                return info
        return answer

    def _extractViewData(self):
        """Extract the data from the current context and build a data structure
        that is usable by the view."""

    def _fieldsToDate(self, year, month, day):
        if not day or not year:
            return None
        return datetime.date(year, month, day)

    @property
    def tree(self):
        return getTreeData(
            self.request, self.context, filter=self.question_filter, phase="actionplan"
        )

    @property
    @memoize
    def number_images(self):
        if self.is_custom_risk:
            return int(bool(self.context.image_filename))
        else:
            number_images = getattr(self.risk, "image", None) and 1 or 0
            if number_images:
                for i in range(2, 5):
                    number_images += getattr(self.risk, f"image{i}", None) and 1 or 0

        return number_images

    def __call__(self):
        super().__call__()
        # Render the page only if the user has inspection rights,
        # otherwise redirect to the start page of the session.
        if not self.webhelpers.can_inspect_session:
            return self.request.response.redirect(
                "{session_url}/@@start".format(
                    session_url=self.webhelpers.traversed_session.absolute_url()
                )
            )
        if self.webhelpers.redirectOnSurveyUpdate():
            return
        context = aq_inner(self.context)
        utils.setLanguage(self.request, self.survey, self.survey.language)

        # already compute "next" here, so that we can know in the template
        # if the next step might be the report phase, in which case we
        # need to switch off the sidebar
        next_question = FindNextQuestion(
            context, dbsession=self.session, filter=self.risk_filter
        )
        if next_question is None:
            # We ran out of questions, proceed to the next phase
            url = "{session_url}/{next_view_name}".format(
                session_url=self.webhelpers.traversed_session.absolute_url(),
                next_view_name=(
                    "@@consultancy"
                    if self.webhelpers.use_consultancy_phase
                    else "@@report"
                ),
            )
        else:
            url = "{session_url}/{path}/@@actionplan".format(
                path="/".join(next_question.short_path),
                session_url=self.webhelpers.traversed_session.absolute_url(),
            )

        previous = FindPreviousQuestion(
            context, dbsession=self.session, filter=self.risk_filter
        )
        if previous is None:
            previous_url = "{session_url}/@@identification".format(
                session_url=self.webhelpers.traversed_session.absolute_url()
            )
        else:
            previous_url = "{session_url}/{path}/@@actionplan".format(
                path="/".join(previous.short_path),
                session_url=self.webhelpers.traversed_session.absolute_url(),
            )

        if self.request.method == "POST":
            reply = self.request.form
            if self.webhelpers.can_edit_session:
                session = Session()
                context.comment = self.webhelpers.get_safe_html(reply.get("comment"))
                context.priority = reply.get("priority")

                new_plans, changes = self.extract_plans_from_request()
                for plan in context.standard_measures + context.custom_measures:
                    session.delete(plan)
                context.action_plans.extend(new_plans)
                if changes:
                    self.session.touch()

            _next = self._get_next(reply)
            if _next == "previous":
                url = previous_url
            return self.request.response.redirect(url)

        self.title = context.parent.title

        if self.is_custom_risk:
            self.risk.description = ""
            self.risk.evaluation_method = ""

        self.image_class = IMAGE_CLASS[self.number_images]
        self.risk_number = self.context.number
        return self.index()


def calculate_priority(db_risk, risk):
    """Update the risk priority.

    This method can be used for risks using a calculated evaluation
    method to determine the priority absed on the subquestions.
    """
    assert risk.evaluation_method == "calculated"
    if risk.type in ["top5", "policy"]:
        db_risk.priority = "high"
    elif evaluation_algorithm(risk) == "french":
        priority = db_risk.frequency * db_risk.effect
        if priority < 10:
            db_risk.priority = "low"
        elif priority <= 45:
            db_risk.priority = "medium"
        else:
            db_risk.priority = "high"
    else:
        priority = db_risk.frequency * db_risk.effect * db_risk.probability
        if priority <= 15:
            db_risk.priority = "low"
        elif priority <= 50:
            db_risk.priority = "medium"
        else:
            db_risk.priority = "high"
    return db_risk.priority


def evaluation_algorithm(risk):
    for parent in aq_chain(aq_inner(risk)):
        if ISurvey.providedBy(parent):
            return getattr(parent, "evaluation_algorithm", "kinney")
    else:
        return "kinney"


class ConfirmationDeleteRisk(BrowserView):
    """View name: @@confirmation-delete-risk."""

    no_splash = True

    @property
    @memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", self.context, self.request)

    @property
    def risk_title(self):
        return self.context.title

    @property
    def risk_id(self):
        return self.context.id

    @property
    def form_action(self):
        return f"{aq_parent(self.context).absolute_url()}/@@delete-risk"

    def __call__(self, *args, **kwargs):
        """Before rendering check if we can find session title."""
        if not self.webhelpers.can_view_session:
            # The user cannot call this view to go the sessions overview.
            return self.request.response.redirect(self.webhelpers.client_url)
        self.risk_title
        return super().__call__(*args, **kwargs)


class DeleteRisk(BrowserView):
    """View name: @@delete-risk."""

    @property
    @memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", self.context, self.request)

    def __call__(self):
        if not self.webhelpers.can_view_session:
            # The user cannot call this view to go the sessions overview.
            return self.request.response.redirect(self.webhelpers.client_url)

        risk_id = self.request.form.get("risk_id", None)
        if risk_id:
            try:
                risk_id = int(risk_id)
            except ValueError:
                pass
            else:
                keep_ids = [
                    risk.id
                    for risk in self.context.children().all()
                    if risk.id != risk_id
                ]
                self.context.removeChildren(excluded=keep_ids)
                notify(CustomRisksModifiedEvent(self.context))

        self.request.response.redirect(
            "{session_url}/@@identification".format(
                session_url=self.context.absolute_url()
            )
        )



from euphorie import MessageFactory as _
from path import Path
from plone import api
from plone.memoize.view import memoize
from Products.Five import BrowserView
from urllib.parse import quote
from weasyprint import CSS
from weasyprint import HTML

import logging


try:
    from tempfile import TemporaryDirectory
except ImportError:
    # PY2
    from backports.tempfile import TemporaryDirectory

log = logging.getLogger(__name__)

# The INFO/WARNING log levels log way to many messages for our needs.
# We have seen 3k log lines for a single pdf conversion in weasyprint 60.2.
# Limit weasyprint's log level to be less chatty.
# This could also be done through configuration:
# https://docs.python.org/3/library/logging.config.html#configuration-file-format
logging.getLogger("weasyprint").setLevel(logging.ERROR)
logging.getLogger("fontTools").setLevel(logging.WARNING)


class PdfView(BrowserView):
    """We use weasyprint to convert an HTML view into a PDF file."""

    @property
    @memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", self.context, self.request)

    @property
    def extra_css(self):
        return """
    @page {{
        @bottom-right {{
            content: "{label_page} " counter(page) " {label_page_of} " counter(pages);
        }}
    }}
        """.format(
            label_page=api.portal.translate(_("label_page", default="Page")),
            label_page_of=api.portal.translate(_("label_page_of", default="of")),
        )

    def view_to_pdf(self, view):
        content = view()
        extra_css = CSS(string=self.extra_css)

        with TemporaryDirectory(prefix="euphoprient") as tmpdir:
            html_file = Path(tmpdir) / "index.html"
            pdf_file = Path(tmpdir) / "index.pdf"
            html_file.write_text(content)
            HTML(html_file).write_pdf(
                pdf_file,
                stylesheets=[extra_css],
            )
            return pdf_file.bytes()

    def __call__(self):
        if not self.webhelpers.can_view_session:
            return self.request.response.redirect(self.webhelpers.client_url)
        context = self.context
        view_name = self.request.get("view", "view")
        view = context.restrictedTraverse(view_name)
        pdf = self.view_to_pdf(view)

        filename = "{} - {}.pdf".format(
            api.portal.translate(view.label), view.session.title
        ).encode("utf-8")

        context.REQUEST.RESPONSE.setHeader("Content-Type", "application/pdf")
        context.REQUEST.RESPONSE.setHeader(
            "Content-Disposition", "attachment;filename*=UTF-8''%s" % quote(filename)
        )
        return pdf



from euphorie.content.behaviour.uniqueid import IIdGenerationRoot
from euphorie.content.behaviour.uniqueid import INameFromUniqueId
from euphorie.content.behaviour.uniqueid import UniqueNameChooser
from zope.annotation.interfaces import IAttributeAnnotatable
from zope.component.testing import PlacelessSetup
from zope.interface import alsoProvides
from zope.interface import implementer

import Acquisition
import unittest


@implementer(INameFromUniqueId)
class Mock(Acquisition.Implicit):
    pass


class IdGenerationTests(PlacelessSetup, unittest.TestCase):
    def setUp(self):
        super().setUp()
        from zope.annotation.attribute import AttributeAnnotations
        from zope.component import provideAdapter

        provideAdapter(AttributeAnnotations)

    def makeRoot(self):
        root = Mock()
        alsoProvides(root, IAttributeAnnotatable, IIdGenerationRoot)
        return root

    def test_ValueErrorIfNoRoot(self):
        chooser = UniqueNameChooser(Mock())
        self.assertRaises(ValueError, chooser.chooseName, None, Mock())

    def testValueErrorIfRootNotAnnotatable(self):
        root = Mock()
        alsoProvides(root, IIdGenerationRoot)
        chooser = UniqueNameChooser(root)
        self.assertRaises(ValueError, chooser.chooseName, None, Mock())

    def testFirstIdIsOne(self):
        root = self.makeRoot()
        obj = Mock()
        UniqueNameChooser(root).chooseName(None, obj)
        self.assertEqual(obj.id, "1")

    def testSecondIdIsTwo(self):
        root = self.makeRoot()
        obj = Mock()
        chooser = UniqueNameChooser(root)
        chooser.chooseName(None, obj)
        del obj.id
        chooser.chooseName(None, obj)
        self.assertEqual(obj.id, "2")

    def testUseExitingIdIfPresent(self):
        root = self.makeRoot()
        obj = Mock()
        obj.id = "mock"
        UniqueNameChooser(root).chooseName(None, obj)
        self.assertEqual(obj.id, "mock")

    def testExtraDepth(self):
        root = self.makeRoot()
        folder = Mock().__of__(root)
        obj = Mock()
        UniqueNameChooser(folder).chooseName(None, obj)
        self.assertEqual(obj.id, "1")



"""
Survey
======

A survey holds the entire risk assessment survey. The survey is a hierarchical
structure containing modules, profile questions and questions. Both modules and
profile questions are used to create the hierarchy.
"""

from .. import MessageFactory as _
from .behaviour.uniqueid import get_next_id
from .behaviour.uniqueid import INameFromUniqueId
from .datamanager import ParentAttributeField
from .fti import check_fti_paste_allowed
from .interfaces import IQuestionContainer
from .profilequestion import IProfileQuestion
from .utils import StripMarkup
from Acquisition import aq_base
from Acquisition import aq_chain
from Acquisition import aq_inner
from Acquisition import aq_parent
from euphorie.content.country import ICountry
from euphorie.content.dependency import ConditionalHtmlText
from euphorie.content.dependency import ConditionalTextLine
from euphorie.content.utils import get_tool_type_default
from euphorie.content.utils import IToolTypesInfo
from euphorie.htmllaundry.z3cform import HtmlText
from plone import api
from plone.app.dexterity.behaviors.metadata import IBasic
from plone.autoform import directives
from plone.dexterity.content import Container
from plone.indexer import indexer
from plone.supermodel import model
from plonetheme.nuplone.z3cform.directives import depends
from plonetheme.nuplone.z3cform.widget import WysiwygFieldWidget
from zope import schema
from zope.component import getUtility
from zope.globalrequest import getRequest
from zope.interface import implementer
from zope.interface import provider
from zope.schema.interfaces import IContextAwareDefaultFactory

import sys


@provider(IContextAwareDefaultFactory)
def _enable_web_training_default(obj):
    """Enable training by default for certain countries.

    The default is enabled if any ancestor country has enabled_web_training set to True.
    When the context has no acquisition chain (happens in tests, for example),
    the default is False.
    """
    # Check if the object is acquired
    if not aq_parent(obj):
        # We are probably in the add form, the object is not yet acquired,
        # Try to get it from the parent country
        request = getRequest()
        parents = request.get("PARENTS", []) or []
        for parent in parents:
            if ICountry.providedBy(parent):
                return parent.enable_web_training
        # If we have no parent country, return False
        return False

    for ancestor in api.content.iter_ancestors(obj, interface=ICountry):
        return ancestor.enable_web_training

    return False


class ISurvey(model.Schema, IBasic):
    """Survey.

    The survey is the root of a survey.
    """

    title = schema.TextLine(
        title=_("label_survey_title", default="Version name"),
        description=_(
            "help_survey_title",
            default="This is the title of this OiRA Tool version. This "
            "name is never shown to users.",
        ),
        required=True,
    )
    directives.order_before(title="*")

    directives.omitted("description")

    introduction = HtmlText(
        title=_("label_introduction", default="Introduction text"),
        description=_(
            "The introduction text is shown when starting a new "
            "OiRA Tool session. If no introduction is provided here a "
            "standard text will be shown. Please keep this text brief "
            "so it will easily fit on screens of small devices such as "
            "phones and PDAs."
        ),
        required=False,
    )
    directives.widget(introduction=WysiwygFieldWidget)

    evaluation_optional = schema.Bool(
        title=_("label_evaluation_optional", default="Evaluation may be skipped"),
        description=_(
            "help_evaluation_optional",
            default="This option allows users to skip the evaluation phase.",
        ),
        default=False,
        required=False,
    )

    language = schema.Choice(
        title=_("label_language", default="Language"),
        vocabulary="plone.app.vocabularies.AvailableContentLanguages",
        default="en",
        required=True,
    )

    classification_code = schema.TextLine(
        title=_("label_classification_code", default="Classification code"),
        description=_(
            "help_classification_code",
            default="A code identifying this sector. Classification "
            "codes are defined by national standards bodies "
            "and based on revision 2 of the NACE standard.",
        ),
        required=False,
    )

    tool_type = schema.Choice(
        title=_("label_tool_type", default="Type of OiRA Tool"),
        description=_(
            "description_tool_type",
            default="This selection determines, which variant of an OiRA Tool"
            ' will be created. If you are not sure, select "Classic".',
        ),
        vocabulary="euphorie.tool_types_vocabulary",
        defaultFactory=get_tool_type_default,
        required=True,
    )

    measures_text_handling = schema.Choice(
        title=_(
            "measures_text_handling",
            default="Handling of measures text (if measures-in-place are used)",
        ),
        description=_(
            "description_measures_text_handling",
            default=(
                "Defines how the “Measures in place” options are displayed "
                "to the user. "
                "Choose “Simple” if the texts for Description and General Approach "
                "are identical in your measures. "
                "Choose “Full” if the texts in General Approach "
                "provide more details than the Description."
            ),
        ),
        vocabulary="euphorie.measures_text_handling_vocabulary",
        default="full",
        required=True,
    )

    integrated_action_plan = schema.Bool(
        title=_("label_integrated_action_plan", default="Integrated Action Plan"),
        description=_(
            "description_integrated_action_plan",
            default="If selected, the option to plan measures will be offered "
            "directly on the “Identification” page. There will be no separate "
            "“Action Plan” step in the navigation.",
        ),
        required=False,
        default=False,
    )

    report_completion_threshold = schema.Int(
        title=_(
            "label_report_completion_threshold",
            default="Completion threshold for report availability",
        ),
        description=_(
            "description_report_completion_threshold",
            default="Please enter the completion percentage above which the report is "
            "available on this tool. The user will be unable to view or download a "
            "report until the assessment is completed to the given percentage. Enter "
            "“0” to always have the report available.",
        ),
        required=True,
        default=0,
        min=0,
        max=100,
    )

    enable_web_training = schema.Bool(
        title=_("label_enable_web_training", default="Enable Web Based Training?"),
        description=_(
            "help_enable_web_training",
            default="If this option is activated, users will be able to take an "
            "online training with this OiRA tool.",
        ),
        required=False,
        defaultFactory=_enable_web_training_default,
    )

    enable_email_reminder = schema.Bool(
        title=_("label_enable_email_reminder", default="Enable email reminder?"),
        description=_(
            "help_enable_email_reminder",
            default="Offer to send an email reminder about this tool",
        ),
        required=False,
    )

    depends("enable_test_questions", "enable_web_training", "on")
    enable_test_questions = schema.Bool(
        title=_(
            "label_enable_test_questions",
            default="Show training questions after the training",
        ),
        description=_(
            "help_enable_test_questions",
            default="If this option is activated, training questions will be shown "
            "during the training. If not activated, questions will only be asked "
            "at the end of the training.",
        ),
        required=False,
        default=False,
    )

    depends("num_training_questions", "enable_test_questions", "on")
    num_training_questions = schema.Int(
        title=_(
            "label_num_training_questions",
            default="Number of questions asked after training",
        ),
        description=_(
            "help_num_training_questions",
            default="Please enter the number of questions to be shown at the end of "
            "your training session. If the number is lower than the number of existing "
            "questions, the questions will be picked randomly. Empty the field to show "
            "all questions.",
        ),
        required=False,
    )

    enable_tool_notification = schema.Bool(
        title=_(
            "label_enable_tool_notification",
            default="Show a custom notification for this OiRA tool?",
        ),
        description=_(
            "description_tool_notification",
            default="If you enter text here, it will be shown to users "
            "in a pop-up when they open the tool. It can be used for "
            "notifying users about changes.",
        ),
        required=False,
        default=False,
    )

    depends("tool_notification_title", "enable_tool_notification", "on")
    tool_notification_title = ConditionalTextLine(
        title=_("label_tool_notification_title", default="Tool notification title"),
        required=True,
    )

    depends("tool_notification_message", "enable_tool_notification", "on")
    tool_notification_message = ConditionalHtmlText(
        title=_("label_tool_notification", default="Tool notification message"),
        required=True,
    )
    directives.widget(tool_notification_message=WysiwygFieldWidget)


class SurveyAttributeField(ParentAttributeField):
    parent_mapping = {
        "survey_title": "title",
        "obsolete": "obsolete",
    }


def get_tool_type(context):
    """Return the tool type used in a given context.

    The type is set on the survey.
    """
    tt_default = get_tool_type_default()
    for parent in aq_chain(aq_inner(context)):
        if ISurvey.providedBy(parent):
            return getattr(parent, "tool_type", "") or tt_default
    return tt_default


@implementer(ISurvey, IQuestionContainer)
class Survey(Container):
    """A risk assessment survey.

    A survey uses the *IIdGenerationRoot* behaviour to guarantee that
    all items inside the survey have a unique id.
    """

    dirty = False

    def _get_id(self, orig_id):
        """Pick an id for pasted content."""
        frame = sys._getframe(1)
        ob = frame.f_locals.get("ob")
        if ob is not None and INameFromUniqueId.providedBy(ob):
            return get_next_id(self)
        return super()._get_id(orig_id)

    def _verifyObjectPaste(self, object, validate_src=True):
        super()._verifyObjectPaste(object, validate_src)
        if validate_src:
            check_fti_paste_allowed(self, object)

    def hasProfile(self):
        """Check if this survey has any profile questions.

        .. todo::    Implement the deprecation checking
        """

        for child in self.values():
            if IProfileQuestion.providedBy(child):
                return True
        else:
            return False

    def hasNotification(self):
        """Checks if a notification message was set."""
        return self.enable_tool_notification or False

    def ProfileQuestions(self):
        """Return a list of all profile questions."""
        return [child for child in self.values() if IProfileQuestion.providedBy(child)]

    def get_tool_type_name(self):
        """Returns the human readable name of the chosen tool type."""
        my_tool_type = get_tool_type(self)
        tti = getUtility(IToolTypesInfo)
        tool_types = tti()
        if my_tool_type not in tool_types:
            my_tool_type = tti.default_tool_type
        return tool_types[my_tool_type]["title"]


@indexer(ISurvey)
def SearchableTextIndexer(obj):
    return " ".join(
        [
            obj.title,
            StripMarkup(obj.description),
            StripMarkup(obj.introduction),
            obj.classification_code or "",
        ]
    )


@indexer(ISurvey)
def LanguageIndexer(obj):
    language = obj.language or ""
    return language.strip().split("-")[0]


class ISurveyAddSchema(model.Schema):
    title = schema.TextLine(
        title=_("label_survey_title", default="Version name"),
        description=_(
            "help_survey_title",
            default="This is the title of this OiRA Tool version. This "
            "name is never shown to users.",
        ),
        required=True,
    )


def handleSurveyUnpublish(survey, event):
    """Event handler (subscriber) for unpublishing a survey."""
    if hasattr(aq_base(survey), "published"):
        delattr(survey, "published")

    surveygroup = aq_parent(survey)
    if surveygroup.published == survey.id:
        surveygroup.published = None



from plone import api
from plone.memoize.view import memoize
from Products.Five import BrowserView
from z3c.saconfig import Session


class BaseView(BrowserView):
    default_target_view = ""

    @property
    @memoize
    def sqlsession(self):
        return Session()

    def redirect(self, target=None, msg="", msg_type="notice"):
        """Redirect the user to a meaningfull place and add a status
        message."""
        if target is None:
            target = self.context.absolute_url()
            if self.default_target_view:
                target = f"{target}/{self.default_target_view}"
        if msg:
            api.portal.show_message(msg, self.request, msg_type)

        self.request.response.redirect(target)

    def handle_POST(self):
        """This is just a hook for subclasses to implement.

        By default do nothing special.
        """
        return super().__call__()

    @property
    @memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", self.context, self.request)

    def validate(self):
        """This is just a hook for subclasses to implement.

        By default do nothing special.
        """
        pass

    def __call__(self):
        self.validate()
        if self.request.method == "POST":
            return self.handle_POST()
        return super().__call__()



from ..sector import getSurveys
from ..sector import ISector
from Acquisition import aq_inner
from euphorie.content import MessageFactory as _
from plone import api
from plone.dexterity.browser.edit import DefaultEditForm
from plonetheme.nuplone.skin import actions
from plonetheme.nuplone.utils import checkPermission
from plonetheme.nuplone.utils import getPortal
from Products.Five import BrowserView
from Products.statusmessages.interfaces import IStatusMessage

import logging
import zExceptions


log = logging.getLogger(__name__)


class SectorView(BrowserView):
    @property
    def add_survey_url(self):
        return "{url}/++add++euphorie.surveygroup".format(
            url=aq_inner(self.context).absolute_url()
        )

    @property
    def surveys(self):
        return getSurveys(self.context)

    @property
    def can_add(self):
        permission = "Euphorie: Add new RIE Content"
        user = api.user.get_current()
        return api.user.has_permission(permission, user=user, obj=self.context)


class EditForm(DefaultEditForm):
    schema = ISector
    default_fieldset_label = None
    formErrorsMessage = "Please correct the indicated errors."

    def extractData(self):
        self.fields = self.fields.omit("title", "login")
        if "title" in self.widgets:
            del self.widgets["title"]
        if "login" in self.widgets:
            del self.widgets["login"]
        return super().extractData()


class VersionCommand(BrowserView):
    def __call__(self):
        action = self.request.get("action")
        if action == "new":
            sector = aq_inner(self.context)
            self.request.response.redirect(
                "%s/++add++euphorie.surveygroup" % sector.absolute_url()
            )
        else:
            log.error("Invalid version command action: %r", action)


class Delete(actions.Delete):
    """Only delete the sector if it doesn't have any published surveys."""

    def verify(self, container, context):
        if not checkPermission(container, "Delete objects"):
            raise zExceptions.Unauthorized

        flash = IStatusMessage(self.request).addStatusMessage
        sector = context
        country = container
        client = getPortal(container).client

        if country.id not in client:
            return True

        cl_country = client[country.id]
        if sector.id not in cl_country:
            return True

        # Look for any published surveys in the client sector, and prevent
        # deletion if any are found
        cl_sector = cl_country[sector.id]
        surveys = [s for s in cl_sector.values() if s.id != "preview"]
        if surveys:
            flash(
                _(
                    "message_not_delete_published_sector",
                    default="You can not delete a sector that contains published "
                    "OiRA Tools.",
                ),
                "error",
            )
            self.request.response.redirect(context.absolute_url())
            return False
        return True



from plone import api
from plone.app.testing.interfaces import SITE_OWNER_NAME


EMPTY_SURVEY = """<sector xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <survey>
              <title>Software development</title>
            </survey>
          </sector>"""


BASIC_SURVEY = """<sector xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <title>ICT</title>
             <survey>
              <title>Software development</title>
              <language>nl</language>
              <module optional="no">
                <title>Module one</title>
                <description>Quick description</description>
                 <risk type="policy">
                   <title>New hires are not aware of design patterns.</title>
                   <description>&lt;p&gt;Every developer should know about them..&lt;/p&gt;</description>
                   <evaluation-method>direct</evaluation-method>
                   <image caption="Key image" content-type="image/gif">R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAEALAAAAAABAAEAAAIBTAA7</image>
                 </risk>
              </module>
            </survey>
          </sector>"""  # noqa: E501


PROFILE_SURVEY = """<sector xmlns="http://xml.simplon.biz/euphorie/survey/1.0">
             <title>ICT</title>
             <survey>
              <title>Software development</title>
              <language>nl</language>
              <profile-question>
                <title>Profile one</title>
                <question>List all your departments:</question>
                <description/>
                <risk type="policy">
                  <title>New hires are not aware of design patterns.</title>
                  <description>&lt;p&gt;Every developer should know about them..&lt;/p&gt;</description>
                  <evaluation-method>direct</evaluation-method>
                </risk>
              </profile-question>
            </survey>
          </sector>"""  # noqa: E501


def _create(container, *args, **kwargs):
    newid = container.invokeFactory(*args, **kwargs)
    obj = getattr(container, newid)
    return obj


def createSector(
    portal,
    id="sector",
    title="Test Sector",
    login=None,
    password=None,
    country="nl",
    **kw
):
    with api.env.adopt_user(SITE_OWNER_NAME):
        if hasattr(portal, "sectors"):
            container = portal.sectors
        else:
            container = _create(portal, "euphorie.sectorcontainer", "sectors")
        if country in container:
            country_obj = container[country]
        else:
            country_obj = _create(container, "euphorie.country", country)

        sector = _create(country_obj, "euphorie.sector", id, title=title, **kw)
        sector.login = login or title.lower()
        sector.password = password if password is not None else sector.login
        return sector


def addSurvey(
    sector,
    snippet=BASIC_SURVEY,
    surveygroup_title="Test survey",
    survey_title="Standard version",
):
    from euphorie.content.browser import upload

    importer = upload.SurveyImporter(sector)
    return importer(snippet, surveygroup_title, survey_title)



"""
Navigation
----------

Assemble data for a navigation tree.
"""

from Acquisition import aq_parent
from euphorie.client import model
from euphorie.client import utils
from euphorie.client.model import SurveySession
from euphorie.content.interfaces import ICustomRisksModule
from euphorie.content.profilequestion import IProfileQuestion
from plone import api
from Products.Five import BrowserView
from sqlalchemy import sql
from z3c.saconfig import Session


def FindFirstQuestion(dbsession, filter=None):
    query = (
        Session.query(model.SurveyTreeItem)
        .filter(model.SurveyTreeItem.session == dbsession)
        .filter(sql.not_(model.SKIPPED_PARENTS))
    )
    if filter is not None:
        query = query.filter(filter)
    return query.order_by(model.SurveyTreeItem.path).first()


def FindNextQuestion(after, dbsession, filter=None):
    query = (
        Session.query(model.SurveyTreeItem)
        .filter(model.SurveyTreeItem.session == dbsession)
        .filter(model.SurveyTreeItem.path > after.path)
        .filter(sql.not_(model.SKIPPED_PARENTS))
    )
    # Skip modules without a description.
    if filter is None:
        filter = model.RISK_OR_MODULE_WITH_DESCRIPTION_FILTER
    else:
        filter = sql.and_(model.RISK_OR_MODULE_WITH_DESCRIPTION_FILTER, filter)
    query = query.filter(filter)
    return query.order_by(model.SurveyTreeItem.path).first()


def FindPreviousQuestion(after, dbsession, filter=None):
    query = (
        Session.query(model.SurveyTreeItem)
        .filter(model.SurveyTreeItem.session == dbsession)
        .filter(model.SurveyTreeItem.path < after.path)
        .filter(sql.not_(model.SKIPPED_PARENTS))
    )
    # Skip modules without a description.
    if filter is None:
        filter = model.RISK_OR_MODULE_WITH_DESCRIPTION_FILTER
    else:
        filter = sql.and_(model.RISK_OR_MODULE_WITH_DESCRIPTION_FILTER, filter)
    query = query.filter(filter)
    return query.order_by(model.SurveyTreeItem.path.desc()).first()


def first(func, iter):
    """Find the first item in an iterable for which func(item) is True.

    If not item is find None is returned.
    """

    for i in iter:
        if func(i):
            return i
    else:
        return None


class TreeDataCreator(BrowserView):
    """Browser view for creating tree data for the navigation."""

    def __call__(
        self,
        element=None,
        phase="identification",
        filter=None,
        survey=None,
        no_current=False,
    ):
        if not survey:
            # Standard, real-world case
            webhelpers = api.content.get_view("webhelpers", self.context, self.request)
            survey = webhelpers._survey
            traversed_session = webhelpers.traversed_session
        else:
            # XXX Fixme
            # Only in tests...
            # In some tests in test_navigation, the view "webhelpers" cannot be found
            # for the given context. That's why we pass in the survey item directly.
            traversed_session = survey

        # This is the tree element that we start from.  It can be the same as the
        # context that gets passed in, if it has an Acquisition chain.
        # On views that are called outside of the context of a module or risk,
        # e.g. the initial @@identification view, the tree-element that we find is not
        # in an acqusition context, so that we cannot use it for fetching the traversed
        # session via webhelpers.
        if not element:
            element = self.context

        query = Session.query(model.SurveyTreeItem)
        title_custom_risks = utils.get_translated_custom_risks_title(self.request)
        root = element
        parents = []
        while root.parent_id is not None:
            parent = query.get(root.parent_id)
            parents.append(parent)
            root = parent
        parents.reverse()

        def morph(obj):
            number = obj.number
            # The custom risks don't have a real number, but an Omega instead
            if obj.zodb_path.find("custom-risks") > -1:
                num_elems = number.split(".")
                number = ".".join(["Ω"] + num_elems[1:])
            info = {
                "id": obj.id,
                "number": number,
                "title": obj.title,
                "active": (
                    obj.path != element.path and element.path.startswith(obj.path)
                ),
                "current": (obj.path == element.path),
                "current_parent": (obj.path == element.path[:-3]),
                "path": element.path,
                "children": [],
                "type": obj.type,
                "leaf_module": False,
                "depth": obj.depth,
                "url": "{session_url}/{obj_path}/@@{phase}".format(
                    session_url=traversed_session.absolute_url(),
                    obj_path="/".join(obj.short_path),
                    phase=phase,
                ),
                "css_id": "",
            }
            cls = []
            for key in ["active", "current", "current_parent"]:
                if info[key]:
                    if key == "current" and no_current:
                        continue
                    cls.append(key)

            if obj.postponed:
                cls.append("postponed")
            else:
                if isinstance(obj, model.Risk):
                    if obj.identification or obj.scaled_answer:
                        cls.append("answered")
                    if obj.identification == "no":
                        cls.append("risk")
                    if obj.scaled_answer:
                        info["scaled_answer"] = obj.scaled_answer

            info["class"] = cls and " ".join(cls) or None
            return info

        # Result is always pointing to the level *above* the current level.
        # At the end it will be the virtual tree root
        result = {
            "children": [],
            "leaf_module": False,
            "current": False,
            "id": None,
            "title": None,
        }
        result["class"] = None
        children = []
        for obj in element.siblings(filter=filter):
            info = morph(obj)
            if obj.type != "risk" and obj.zodb_path.find("custom-risks") > -1:
                info["title"] = title_custom_risks
                info["css_id"] = "other-risks"
            children.append(info)
        result["children"] = children

        if isinstance(element, model.Module):
            # If this is an optional module, check the "postponed" flag.
            # As long as the optional question has not been answered, skip
            # showing its children.
            # Only a "Yes" answer will set skip_children to False
            module = survey.restrictedTraverse(element.zodb_path.split("/"))
            # In the custom risks module, we never skip children
            # Due to historical reasons, some custom modules might be set to
            # postponed. Here, we ignore that setting.
            if ICustomRisksModule.providedBy(module):
                element.skip_children = False
            elif getattr(module, "optional", False) and element.postponed in (
                True,
                None,
            ):
                element.skip_children = True
            if not element.skip_children:
                # For modules which do not skip children, include the list of
                # children.
                me = first(lambda x: x["current"], result["children"])
                children = []
                for obj in element.children(filter=filter):
                    info = morph(obj)
                    # XXX: The check for SurveySession is due to Euphorie tests which
                    # don't have a proper canonical ZODB survey object and don't test
                    # the following OiRA-specific code.
                    if (
                        obj.depth == 2
                        and not getattr(obj, "is_custom_risk", False)
                        and not isinstance(survey, SurveySession)
                    ):
                        module = survey.restrictedTraverse(obj.zodb_path.split("/"))
                        if IProfileQuestion.providedBy(
                            module
                        ) and not ICustomRisksModule.providedBy(aq_parent(module)):
                            info["type"] = "location"
                            info["children"] = [
                                morph(sub) for sub in obj.children(filter=filter)
                            ]
                    children.append(info)
                me["children"] = children
                types = {c["type"] for c in me["children"]}
                me["leaf_module"] = "risk" in types

        elif isinstance(element, model.Risk):
            # For a risk we also want to include all siblings of its module parent
            parent = parents.pop()
            siblings = []
            for obj in parent.siblings(model.Module, filter=filter):
                info = morph(obj)
                if obj.zodb_path.find("custom-risks") > -1:
                    info["title"] = title_custom_risks
                    info["css_id"] = "other-risks"
                siblings.append(info)
            myparent = first(lambda x: x["active"], siblings)
            myparent["children"] = result["children"]
            myparent["leaf_module"] = True
            result["children"] = siblings

        if parents:
            # Add all parents up to the root
            while len(parents) > 1:
                parent = parents.pop()
                new = morph(parent)
                siblings = []
                if isinstance(parent, model.Module) and parent.depth == 2:
                    module = survey.restrictedTraverse(parent.zodb_path.split("/"))
                    if IProfileQuestion.providedBy(
                        module
                    ) and not ICustomRisksModule.providedBy(aq_parent(module)):
                        new["type"] = "location"
                        # Include the siblings and the first level of the sibling trees
                        # so that they are still visible and browsable from the sidebar.
                        for sibling in parent.siblings(model.Module, filter=filter):
                            if sibling.id == new["id"]:
                                new["children"] = result["children"]
                                siblings.append(new)
                            else:
                                info = morph(sibling)
                                info["type"] = "location"
                                children = []
                                for child in sibling.children(filter=filter):
                                    child_info = morph(child)
                                    children.append(child_info)
                                info["children"] = children
                                siblings.append(info)
                if not siblings:
                    new["children"] = result["children"]
                    siblings.append(new)
                result["children"] = siblings

            # Finally list all modules at the root level
            parent = parents.pop()
            roots = []
            for obj in parent.siblings(model.Module, filter=filter):
                info = morph(obj)
                if obj.zodb_path.find("custom-risks") > -1:
                    info["title"] = title_custom_risks
                roots.append(info)

            myroot = first(lambda x: x["active"], roots)
            myroot["children"] = result["children"]
            result["children"] = roots

        return result


def getTreeData(
    request,
    context,
    element=None,
    phase="identification",
    filter=None,
    survey=None,
    no_current=False,
):
    """Assemble data for a navigation tree.

    This function returns a nested dictionary structure reflecting the
    elements for a navigation tree. The tree will all sibling questions of
    the current context, the current module and all its module siblings, its
    parents up to the root module, and all modules at the root level.

    Optionally a SQLAlchemy clause can be provided, which will be used to
    filter items shown in the tree. The current item and its parents will
    always be shown.

    Each element is reflect as a dictionary item with the following keys:

    - id: the SQL object id
    - type: the SQL object type
    - number: a human presentable numbering of the item
    - title: the object title
    - current: boolean indicating if this is the current context or its
      direct parent module
    - active: boolean indicating if this is a parent node of the current
      context
    - class: CSS classes to use for this node
    - children: a list of child nodes (in the right order)
    - url: URL for this item

    We now put all logic into a class, making it more easily overridable.
    """
    try:
        tree_data_creator = api.content.get_view(
            "oira_navigation_tree", context, request
        )
    except api.exc.InvalidParameterError:
        # XXX Fixme
        # In the GetTreeDataTests in test_navigation, the view "oira_navigation_tree"
        # cannot be found.   Same is true for the "webhelpers" view, which we call in
        # the "oira_navigation_tree" view.  In fact only 8 views are found.
        # So test setup could be improved, but that has been the case for a while.
        tree_data_creator = TreeDataCreator(context, request)

    tree_data = tree_data_creator(
        element=element,
        phase=phase,
        filter=filter,
        survey=survey,
        no_current=no_current,
    )
    return tree_data



from euphorie.testing import EuphorieFunctionalTestCase
from euphorie.testing import EuphorieIntegrationTestCase
from plone import api
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from transaction import commit


class ModuleTests(EuphorieIntegrationTestCase):
    def _create(self, container, *args, **kwargs):
        newid = container.invokeFactory(*args, **kwargs)
        return getattr(container, newid)

    def createModule(self):
        country = self.portal.sectors.nl
        sector = self._create(country, "euphorie.sector", "sector")
        surveygroup = self._create(sector, "euphorie.surveygroup", "group")
        survey = self._create(surveygroup, "euphorie.survey", "survey")
        module = self._create(survey, "euphorie.module", "module")
        return module

    def testNotGloballyAllowed(self):
        self.loginAsPortalOwner()
        types = [fti.id for fti in self.portal.allowedContentTypes()]
        self.assertTrue("euphorie.module" not in types)

    def testAllowedContentTypes(self):
        self.loginAsPortalOwner()
        module = self.createModule()
        types = [fti.id for fti in module.allowedContentTypes()]
        self.assertEqual(set(types), {"euphorie.module", "euphorie.risk"})

    def testConditionalFtiUsed(self):
        from euphorie.content.fti import ConditionalDexterityFTI

        fti = getattr(self.portal.portal_types, "euphorie.module")
        self.assertTrue(isinstance(fti, ConditionalDexterityFTI))

    def testCanBeCopied(self):
        self.loginAsPortalOwner()
        module = self.createModule()
        self.assertTrue(module.cb_isCopyable())

    def test_verifyObjectPaste_acceptablePaste(self):
        from Acquisition import aq_parent

        self.loginAsPortalOwner()
        target = self.createModule()
        survey = aq_parent(target)
        source = self._create(survey, "euphorie.module", "other")
        target._verifyObjectPaste(source)

    def test_verifyObjectPaste_block_if_result_too_deep(self):
        from Acquisition import aq_parent

        self.loginAsPortalOwner()
        target = self.createModule()
        survey = aq_parent(target)
        source = self._create(survey, "euphorie.module", "other")
        other = self._create(source, "euphorie.module", "other")
        self._create(other, "euphorie.module", "other")
        self.assertRaises(ValueError, target._verifyObjectPaste, source)


class ConstructionFilterTests(EuphorieIntegrationTestCase):
    def _create(self, container, *args, **kwargs):
        newid = container.invokeFactory(*args, **kwargs)
        return getattr(container, newid)

    def createStructure(self):
        self.country = self.portal.sectors.nl
        self.sector = self._create(self.country, "euphorie.sector", "sector")
        self.surveygroup = self._create(self.sector, "euphorie.surveygroup", "group")
        self.survey = self._create(self.surveygroup, "euphorie.survey", "survey")
        self.module = self._create(self.survey, "euphorie.module", "module")

    def testValidDepthInModule(self):
        self.loginAsPortalOwner()
        self.createStructure()
        types = [fti.id for fti in self.module.allowedContentTypes()]
        self.assertTrue("euphorie.module" in types)

    def testMaxDepthInModule(self):
        self.loginAsPortalOwner()
        self.createStructure()
        submodule = self._create(self.module, "euphorie.module", "module")
        subsubmodule = self._create(submodule, "euphorie.module", "module")
        types = [fti.id for fti in subsubmodule.allowedContentTypes()]
        self.assertTrue("euphorie.module" not in types)

    def testPreventModuleIfRiskExists(self):
        self.loginAsPortalOwner()
        self.createStructure()
        types = [fti.id for fti in self.module.allowedContentTypes()]
        self.assertTrue("euphorie.module" in types)
        self._create(self.module, "euphorie.risk", "risk")
        types = [fti.id for fti in self.module.allowedContentTypes()]
        self.assertTrue("euphorie.module" not in types)


class FunctionalTests(EuphorieFunctionalTestCase):
    _default_credentials = {
        "username": SITE_OWNER_NAME,
        "password": SITE_OWNER_PASSWORD,
    }

    def _create(self, container, *args, **kwargs):
        with api.env.adopt_user("admin"):
            newid = container.invokeFactory(*args, **kwargs)
            return getattr(container, newid)

    def createStructure(self):
        with api.env.adopt_user("admin"):
            self.country = self.portal.sectors.nl
            self.sector = self._create(self.country, "euphorie.sector", "sector")
            self.surveygroup = self._create(
                self.sector, "euphorie.surveygroup", "group"
            )
            self.survey = self._create(self.surveygroup, "euphorie.survey", "survey")
            self.module = self._create(self.survey, "euphorie.module", "module")
            commit()

    def testEditTitleForModule(self):
        self.createStructure()
        browser = self.get_browser(logged_in=True)
        browser.open("%s/@@edit" % self.module.absolute_url())
        self.assertTrue("Edit Module" in browser.contents)

    def testEditTitleForSubModule(self):
        self.createStructure()
        submodule = self._create(self.module, "euphorie.module", "module")
        browser = self.get_browser(logged_in=True)
        browser.open("%s/@@edit" % submodule.absolute_url())
        self.assertTrue("Edit Module" not in browser.contents)
        self.assertTrue("Edit Submodule" in browser.contents)

    def testInactiveHideFromTrainingBehavior(self):
        """When `euphorie.hide_from_training` is not activated, the
        `hide_from_training` field should not be visible.
        """
        self.createStructure()

        browser = self.get_browser(logged_in=True)
        browser.open("%s/@@edit" % self.module.absolute_url())

        self.assertTrue(
            'name="form.widgets.IHideFromTraining.hide_from_training:list"'
            not in browser.contents
        )

    def testActiveHideFromTrainingBehavior(self):
        """When `euphorie.hide_from_training` is activated, the
        `hide_from_training` field should be visible.
        """
        self.createStructure()

        portal_types = api.portal.get_tool("portal_types")
        fti = portal_types["euphorie.module"]
        fti.behaviors = tuple(list(fti.behaviors) + ["euphorie.hide_from_training"])

        browser = self.get_browser(logged_in=True)
        browser.open("%s/@@edit" % self.module.absolute_url())

        self.assertTrue(
            'name="form.widgets.IHideFromTraining.hide_from_training:list"'
            in browser.contents
        )



from .. import MessageFactory as _
from euphorie.content.behaviour.richdescription import IRichDescription
from euphorie.content.utils import StripMarkup
from euphorie.htmllaundry.z3cform import HtmlText
from plone.app.dexterity.behaviors.metadata import IBasic
from plone.autoform import directives
from plone.indexer import indexer
from plone.supermodel import model
from plonetheme.nuplone.z3cform.widget import WysiwygFieldWidget


class IPage(model.Schema, IRichDescription, IBasic):
    """A basic page."""

    description = HtmlText(
        title=_("label_module_description", "Description"),
        description=_(
            "help_module_description",
            default="Include any relevant information that may be "
            "helpful for the end-user.",
        ),
        required=True,
    )
    directives.widget(description=WysiwygFieldWidget)
    directives.order_after(description="title")

    body = HtmlText(title=_("label_body", "Page content"), required=True)
    directives.widget(body=WysiwygFieldWidget)


@indexer(IPage)
def SearchableTextIndexer(obj):
    return " ".join([obj.title, StripMarkup(obj.description), StripMarkup(obj.body)])



from euphorie.client.model import Risk
from euphorie.client.model import Session
from ftw.upgrade import UpgradeStep
from logging import getLogger

import sqlalchemy as sa


logger = getLogger()


class DecommissionTheTrainingNotesField(UpgradeStep):
    """Decommission the training_notes field.

    For the moment we are not going to remove the field from the
    database, but we copy the values to the comment column if it has no
    entry.
    """

    def __call__(self):
        session = Session()
        query = session.query(Risk).filter(
            sa.and_(Risk.training_notes != "", sa.func.coalesce(Risk.comment, "") == "")
        )
        for risk in query:
            logger.info("Copying training_notes to comment for risk %r", risk.id)
            risk.comment = risk.training_notes



from datetime import datetime
from euphorie.client import model
from euphorie.client.tests.utils import addAccount
from euphorie.client.tests.utils import addSurvey
from euphorie.content.tests.utils import BASIC_SURVEY
from euphorie.testing import EuphorieIntegrationTestCase
from plone import api


class TestCloningViews(EuphorieIntegrationTestCase):
    def setUp(self):
        super().setUp()
        self.loginAsPortalOwner()
        addSurvey(self.portal, BASIC_SURVEY)
        self.jane = addAccount("jane@example.com", password="secret")
        self.john = addAccount("john@example.com", password="secret")

        group = model.Group(group_id="1")
        model.Session.add(group)

        self.jane.group = group
        self.john.group = group
        model.Session.flush()

        survey_session = model.SurveySession(
            title="Dummy session",
            created=datetime(2012, 4, 22, 23, 5, 12),
            modified=datetime(2012, 4, 23, 11, 50, 30),
            zodb_path="nl/ict/software-development",
            account=self.jane,
            group=group,
            company=model.Company(country="nl", employees="1-9", referer="other"),
        )
        module = survey_session.addChild(
            model.Module(title="module 1", module_id="1", zodb_path="a")
        )
        risk = module.addChild(
            model.Risk(title="question 1", risk_id="1", zodb_path="a/b")
        )
        model.ActionPlan(action_plan="This is the plan", risk=risk)
        model.Session.add(survey_session)

    def test_clone(self):
        country = self.portal.client.nl
        traversed_session = country.ict["software-development"].restrictedTraverse(
            "++session++1"
        )
        # Calling the view redirects to the client home page if we are not autenticated
        with self._get_view("clone-session", traversed_session) as view:
            view()
            self.assertDictEqual(
                view.request.response.headers,
                {"location": "http://nohost/plone/client"},
            )

        # Calling the view redirects to the cloned session if authenticated
        with api.env.adopt_user(user=self.john):
            with self._get_view(
                "clone-session", traversed_session, traversed_session.session
            ) as view:
                view()
                self.assertDictEqual(
                    view.request.response.headers,
                    {
                        "location": "http://nohost/plone/client/nl/ict/software-development/++session++2/@@start?new_clone=1"  # noqa: E501
                    },
                )

        with api.env.adopt_user(user=self.john):
            with self._get_view("clone-session", traversed_session) as view:
                original = model.Session.query(model.SurveySession).get(1)
                clone = view.get_cloned_session()
                # The cloned session has a lot in common with the original one
                # The title is prefixed with a COPY marker
                self.assertEqual(clone.title, f"COPY: {original.title}")
                self.assertEqual(clone.company.country, original.company.country)
                # But is marked with the account that made the clone
                self.assertEqual(original.account.login, "jane@example.com")
                self.assertEqual(clone.account.login, "john@example.com")

                # Also all the tree is copied
                self.assertListEqual(
                    [(x.id, x.title) for x in clone.children()], [(5, "module 1")]
                )
                module = clone.children()[0]
                self.assertListEqual(
                    [(x.id, x.title) for x in module.children()], [(6, "question 1")]
                )
                risk = module.children()[0]
                self.assertListEqual(
                    [(x.id, x.action_plan) for x in risk.action_plans],
                    [(3, "This is the plan")],
                )
                # Also verify that the cloned risk has
                # the cloned module as a parent
                self.assertEqual(risk.parent, module)
